# coding: utf-8
"""
Utilities for saving and loading predictions
"""
import os
import sys
import traceback as tb
import numpy as np

from .pretty_print import print_warning


def save_y_pred(problem, y_pred, data_path='.', output_path='.',
                suffix='test'):
    """Save a prediction vector in file.

    If problem.save_y_pred is implemented, y_pred is passed to it. Otherwise,
    np.savez_compressed is used on y_pred. If it crashes, a warning is raised.
    The file is (typically) in
    submissions/<submission_name>/training_output/y_pred_<suffix>.npz or
    submissions/<submission_name>/training_output/fold_<i>/y_pred_<suffix>.npz.

    Parameters
    ----------
    problem : a problem object
        loaded from problem.py, may implement save_y_pred
    y_pred : a prediction vector
        a vector of predictions to be saved
    data_path : str, (default='.')
        the directory of the ramp-kit to be tested for submission, maybe
        needed by problem.save_y_pred for, e.g., merging with an index vector
    output_path : str, (default='.')
        the directory where (typically) y_pred_<suffix>.npz will be saved
    suffix : str, (default='test')
        suffix in (typically) y_pred_<suffix>.npz, can be used in
        problem.save_y_pred to, e.g., save only test predictions
    """
    try:
        # We try using custom made problem.save_y_pred
        # obligatory to implement if np.savez_compressed doesn't work on y_pred
        problem.save_y_pred(y_pred, data_path, output_path, suffix)
    except AttributeError:
        # We fall back to numpy savez_compressed
        try:
            y_pred_f_name = os.path.join(output_path,
                                         'y_pred_{}'.format(suffix))
            np.savez_compressed(y_pred_f_name, y_pred=y_pred)
        except Exception as e:
            print_warning(
                "Warning: model can't be saved.\n{}\n".format(e) +
                'Consider implementing custom save_y_pred in problem.py\n')


def load_y_pred(problem, data_path='.', input_path='.', suffix='test'):
    """Load a file into a prediction vector.

    If problem.load_y_pred is implemented, y_pred is loaded by it. Otherwise,
    np.load is used. If it crashes, the exception is raised. The input file
    is (typically) in
    submissions/<submission_name>/training_output/y_pred_<suffix>.npz or
    submissions/<submission_name>/training_output/fold_<i>/y_pred_<suffix>.npz.

    Parameters
    ----------
    problem : a problem object
        loaded from problem.py, may implement save_y_pred
    data_path : str, (default='.')
        the directory of the ramp-kit to be tested for submission, maybe
        needed by problem.save_y_pred for, e.g., merging with an index vector
    input_path : str, (default='.')
        the directory where (typically) y_pred_<suffix>.npz will be saved
    suffix : str, (default='test')
        suffix in (typically) y_pred_<suffix>.npz
    """
    try:
        # We try using custom made problem.load_y_pred
        # obligatory to implement if np.load doesn't work
        return problem.load_y_pred(data_path, input_path, suffix)
    except AttributeError:
        # We fall back to numpy load
        y_pred_f_name = os.path.join(input_path,
                                     'y_pred_{}.npz'.format(suffix))
        return np.load(y_pred_f_name)['y_pred']


def set_state(state, save_output, output_path):
    """Save the submission state (per fold) in <output_path>/state.txt.

    In case save_output is True. Otherwise do nothing.

    Parameters
    ----------
    state : str
        The string representing the state. Possible values are 'new',
        'trained', 'validated', 'tested', 'scored', 'training_error',
        'validating_error', 'testing_error'.
    save_output : boolean
        True if state should be written in file
    output_path : str
        the path into which 'state.txt' will be saved
    """
    if save_output:
        with open(os.path.join(output_path, 'state.txt'), 'w') as fd:
            fd.write(state)


def print_submission_exception(save_output, output_path):
    """Print the exception trace corresponding the user submission.

    In case save_output is True, also save it into <output_path>/error.txt

    Parameters
    ----------
    save_output : boolean
        True if error should be written in file
    output_path : str
        the path into which 'error.txt' will be saved
    """
    exc_type, exc_value, exc_traceback = sys.exc_info()
    # print the stack corresponding to user submission
    trace = tb.format_exception(exc_type, exc_value, exc_traceback)
    for s in trace:
        print(s)
    # some times traces is <5
    try:
        if trace[4].find('load_source') > -1:
            trace = trace[5:]
    except IndexError:
        trace = trace[4:]
    if save_output:
        with open(os.path.join(output_path, 'error.txt'), 'w') as fd:
            for s in trace:
                fd.write(s)



import os

from ..utils.importing import import_module_from_source


class Regressor(object):
    def __init__(self, workflow_element_names=['regressor']):
        self.element_names = workflow_element_names
        # self.name = 'regressor_workflow'  # temporary

    def train_submission(self, module_path, X_array, y_array, train_is=None):
        if train_is is None:
            train_is = slice(None, None, None)
        regressor = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        reg = regressor.Regressor()
        reg.fit(X_array[train_is], y_array[train_is])
        return reg

    def test_submission(self, trained_model, X_array):
        reg = trained_model
        y_pred = reg.predict(X_array)
        return y_pred



from __future__ import division

import os

import numpy as np

from ..utils.importing import import_module_from_source


class ImageClassifier(object):
    """
    ImageClassifier workflow.

    This workflow is used to train image classification tasks, typically when
    the dataset cannot be stored in memory.
    Submissions need to contain two files, which by default are named:
    image_preprocessor.py and batch_classifier.py (they can be modified
    by changing `workflow_element_names`).

    image_preprocessor.py needs a `tranform` function, which
    is used for preprocessing the images. It takes an image as input
    and it returns an image as an output. Optionally, image_preprocessor.py
    can also have a function `transform_test`, which is used only to preprocess
    images at test time. Otherwise, if `transform_test` does not exist,
    `transform` is used at train and test time.

    batch_classifier.py needs a `BatchClassifier` class, which implements
    `fit` and `predict_proba`, where `fit` takes as input an instance
    of `BatchGeneratorBuilder`.

    Parameters
    ==========

    test_batch_size : int
        batch size used for testing.

    chunk_size : int
        size of the chunk used to load data from disk into memory.
        (see at the top of the file what a chunk is and its difference
         with the mini-batch size of neural nets).

    n_jobs : int
        the number of jobs used to load images from disk to memory as `chunks`.

    n_classes : int
        Total number of classes.

    """

    def __init__(self, test_batch_size, chunk_size, n_jobs, n_classes,
                 workflow_element_names=[
                     'image_preprocessor', 'batch_classifier']):
        self.element_names = workflow_element_names
        self.test_batch_size = test_batch_size
        self.chunk_size = chunk_size
        self.n_jobs = n_jobs
        self.n_classes = n_classes

    def train_submission(self, module_path, folder_X_array, y_array,
                         train_is=None):
        """Train a batch image classifier.

        module_path : str
            module where the submission is. the folder of the module
            have to contain batch_classifier.py and image_preprocessor.py.
        X_array : ArrayContainer vector of int
            vector of image IDs to train on
            (it is named X_array to be coherent with the current API,
             but as said here, it does not represent the data itself,
             only image IDs).
        y_array : vector of int
            vector of image labels corresponding to X_train
        train_is : vector of int
           indices from X_array to train on
        """
        folder, X_array = folder_X_array
        if train_is is None:
            train_is = slice(None, None, None)
        image_preprocessor = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        transform_img = image_preprocessor.transform
        transform_test_img = getattr(image_preprocessor,
                                     'transform_test',
                                     transform_img)
        batch_classifier = import_module_from_source(
            os.path.join(module_path, self.element_names[1] + '.py'),
            self.element_names[1],
            sanitize=True
        )
        clf = batch_classifier.BatchClassifier()

        gen_builder = BatchGeneratorBuilder(
            X_array[train_is], y_array[train_is],
            transform_img, transform_test_img,
            folder=folder,
            chunk_size=self.chunk_size, n_classes=self.n_classes,
            n_jobs=self.n_jobs)
        clf.fit(gen_builder)
        return transform_img, transform_test_img, clf

    def test_submission(self, trained_model, folder_X_array):
        """Train a batch image classifier.

        trained_model : tuple (function, Classifier)
            tuple of a trained model returned by `train_submission`.
        X_array : ArrayContainer of int
            vector of image IDs to test on.
            (it is named X_array to be coherent with the current API,
             but as said here, it does not represent the data itself,
             only image IDs).
        """
        folder, X_array = folder_X_array
        transform_img, transform_test_img, clf = trained_model
        it = _chunk_iterator(
            X_array, folder=folder, chunk_size=self.chunk_size)
        y_proba = []
        for X in it:
            for i in range(0, len(X), self.test_batch_size):
                # 1) Preprocessing
                X_batch = X[i: i + self.test_batch_size]
                # X_batch = Parallel(n_jobs=self.n_jobs, backend='threading')(
                #     delayed(transform_img)(x) for x in X_batch)
                X_batch = [transform_test_img(x) for x in X_batch]
                # X is a list of numpy arrrays at this point, convert it to a
                # single numpy array.
                try:
                    X_batch = [x[np.newaxis, :, :, :] for x in X_batch]
                except IndexError:
                    # single channel
                    X_batch = [
                        x[np.newaxis, np.newaxis, :, :] for x in X_batch]
                X_batch = np.concatenate(X_batch, axis=0)

                # 2) Prediction
                y_proba_batch = clf.predict_proba(X_batch)
                y_proba.append(y_proba_batch)
        y_proba = np.concatenate(y_proba, axis=0)
        return y_proba


class BatchGeneratorBuilder(object):
    """A batch generator builder for generating images on the fly.

    This class is a way to build training and
    validation generators that yield each time a tuple (X, y) of mini-batches.
    The generators are built in a way to fit into keras API of `fit_generator`
    (see https://keras.io/models/model/).
    An instance of this class is exposed to users `Classifier` through
    the `fit` function : model fitting is called by using
    "clf.fit(gen_builder)" where `gen_builder` is an instance
    of this class : `BatchGeneratorBuilder`.
    The fit function from `Classifier` should then use the instance
    to build train and validation generators, using the method
    `get_train_valid_generators`

    Parameters
    ==========

    X_array : ArrayContainer of int
        vector of image IDs to train on
         (it is named X_array to be coherent with the current API,
         but as said here, it does not represent the data itself,
         only image IDs).

    y_array : vector of int
        vector of image labels corresponding to `X_array`

    folder : str
        folder where the images are

    chunk_size : int
        size of the chunk used to load data from disk into memory.
        (see at the top of the file what a chunk is and its difference
         with the mini-batch size of neural nets).

    n_classes : int
        Total number of classes. This is needed because the array
        of labels, which is a vector of ints, is transformed into
        a onehot representation.

    n_jobs : int
        the number of jobs used to load images from disk to memory as `chunks`.
    """

    def __init__(self, X_array, y_array,
                 transform_img, transform_test_img,
                 folder, chunk_size, n_classes, n_jobs):
        self.X_array = X_array
        self.y_array = y_array
        self.transform_img = transform_img
        self.transform_test_img = transform_test_img
        self.folder = folder
        self.chunk_size = chunk_size
        self.n_classes = n_classes
        self.n_jobs = n_jobs
        self.nb_examples = len(X_array)

    def get_train_valid_generators(self, batch_size=256, valid_ratio=0.1):
        """Build train and valid generators for keras.

        This method is used by the user defined `Classifier` to o build train
        and valid generators that will be used in keras `fit_generator`.

        Parameters
        ==========

        batch_size : int
            size of mini-batches
        valid_ratio : float between 0 and 1
            ratio of validation data

        Returns
        =======

        a 4-tuple (gen_train, gen_valid, nb_train, nb_valid) where:
            - gen_train is a generator function for training data
            - gen_valid is a generator function for valid data
            - nb_train is the number of training examples
            - nb_valid is the number of validation examples
        The number of training and validation data are necessary
        so that we can use the keras method `fit_generator`.
        """
        nb_valid = int(valid_ratio * self.nb_examples)
        nb_train = self.nb_examples - nb_valid
        indices = np.arange(self.nb_examples)
        train_indices = indices[0:nb_train]
        valid_indices = indices[nb_train:]
        gen_train = self._get_generator(
            indices=train_indices, batch_size=batch_size)
        gen_valid = self._get_generator(
            indices=valid_indices, batch_size=batch_size)
        return gen_train, gen_valid, nb_train, nb_valid

    def _get_generator(self, indices=None, batch_size=256):
        if indices is None:
            indices = np.arange(self.nb_examples)
        # Infinite loop, as required by keras `fit_generator`.
        # However, as we provide the number of examples per epoch
        # and the user specifies the total number of epochs, it will
        # be able to end.
        while True:
            it = _chunk_iterator(
                X_array=self.X_array[indices], folder=self.folder,
                y_array=self.y_array[indices], chunk_size=self.chunk_size,
                n_jobs=self.n_jobs)
            for X, y in it:
                # 1) Preprocessing of X and y
                # X = Parallel(
                # n_jobs=self.n_jobs, backend='threading')(delayed(
                #     self.transform_img)(x) for x in X)
                X = np.array([self.transform_img(x) for x in X])
                # # X is a list of numpy arrrays at this point, convert it to a
                # single numpy array.
                try:
                    X = [x[np.newaxis, :, :, :] for x in X]
                except IndexError:
                    # single channel
                    X = [x[np.newaxis, np.newaxis, :, :] for x in X]
                X = np.concatenate(X, axis=0)
                X = np.array(X, dtype='float32')
                # Convert y to onehot representation
                y = _to_categorical(y, num_classes=self.n_classes)

                # 2) Yielding mini-batches
                for i in range(0, len(X), batch_size):
                    yield X[i:i + batch_size], y[i:i + batch_size]


def _chunk_iterator(X_array, folder, y_array=None, chunk_size=1024, n_jobs=8):
    """Generate chunks of images, optionally with their labels.

    Parameters
    ==========

    X_array : ArrayContainer of int
        image ids to load
        (it is named X_array to be coherent with the current API,
         but as said here, it does not represent the data itself,
         only image IDs).

    y_array : vector of int
        labels corresponding to each image from X_array

    chunk_size : int
        chunk size

    folder : str
        folder where the images are

    n_jobs : int
        number of jobs used to load images in parallel

    Yields
    ======

    if y_array is provided (not None):
        it yields each time a tuple (X, y) where X is a list
        of numpy arrays of images and y is a list of ints (labels).
        The length of X and y is `chunk_size` at most (it can be smaller).

    if y_array is not provided (it is None)
        it yields each time X where X is a list of numpy arrays
        of images. The length of X is `chunk_size` at most (it can be
        smaller).
        This is used for testing, where we don't have/need the labels.

    The shape of each element of X in both cases
    is (height, width, color), where color is 1 or 3 or 4 and height/width
    vary according to examples (hence the fact that X is a list instead of
    numpy array).
    """
    from skimage.io import imread
    from joblib import delayed
    from joblib import Parallel
    for i in range(0, len(X_array), chunk_size):
        X_chunk = X_array[i:i + chunk_size]
        filenames = [
            os.path.join(folder, '{}'.format(x))
            for x in X_chunk]
        X = Parallel(n_jobs=n_jobs, backend='threading')(delayed(imread)(
            filename) for filename in filenames)
        if y_array is not None:
            y = y_array[i:i + chunk_size]
            yield X, y
        else:
            yield X


def _to_categorical(y, num_classes=None):
    """Convert a class vector (integers) to binary class matrix.

    Taken from keras:
    https://github.com/fchollet/keras/blob/master/keras/utils/np_utils.py
    The reason it was taken from keras is to avoid importing theano which
    clashes with pytorch.

    E.g. for use with categorical_crossentropy.
    # Arguments
        y: class vector to be converted into a matrix
            (integers from 0 to num_classes).
        num_classes: total number of classes.
    # Returns
        A binary matrix representation of the input.
    """
    y = np.array(y, dtype='int').ravel()
    if not num_classes:
        num_classes = np.max(y) + 1
    n = y.shape[0]
    categorical = np.zeros((n, num_classes))
    categorical[np.arange(n), y] = 1
    return categorical


def get_nb_minibatches(nb_samples, batch_size):
    """Compute the number of minibatches for keras.

    See [https://keras.io/models/sequential]
    """
    return (nb_samples // batch_size) +\
        (1 if (nb_samples % batch_size) > 0 else 0)



"""Testing for detection predictions (rampwf.prediction.detection)."""

import numpy as np
from numpy.testing import assert_array_equal, assert_allclose

from rampwf.prediction_types.detection import Predictions


def test_combine():

    pred1 = Predictions(
        y_pred=[[(1, 1, 1, 1), (0.65, 3, 3, 1)]])
    pred2 = Predictions(
        y_pred=[[(0.9, 1.1, 1.1, 1.1), (0.5, 3, 3, 1), (0.8, 5, 5, 1)]])
    pred3 = Predictions(
        y_pred=[[(0.7, 3, 3, 1), (0.6, 5.1, 5.1, 1), (0.3, 10, 10, 1)]])

    y_pred_combined = Predictions.combine([pred1, pred2, pred3]).y_pred

    assert y_pred_combined.shape[0] == 1
    assert len(y_pred_combined[0]) == 3


def test_combine_no_match():

    pred1 = Predictions(
        y_pred=[[], [(1, 1, 1, 1)]])
    pred2 = Predictions(
        y_pred=[[], [(1, 3, 3, 1)]])

    y_pred_combined = Predictions.combine([pred1, pred2]).y_pred

    expected = np.empty(2, dtype=object)
    expected[:] = [[], []]

    assert_array_equal(expected, y_pred_combined)


def test_combine_ignore_none():

    pred1 = Predictions(
        y_pred=[[(1., 1, 1, 1)], None])
    pred2 = Predictions(
        y_pred=[[(1., 1, 1, 1)], [(1., 3, 3, 1)]])
    pred3 = Predictions(
        y_pred=[[(1., 3, 3, 1)], [(1., 3, 3, 1)]])

    y_pred_combined = Predictions.combine([pred1, pred2, pred3]).y_pred

    # first item: 2 out of 3 match -> confidence of 2/3
    # second item: ignore None, so 2 out of 2 match -> confidence of 2/2 = 1
    expected = np.empty(2, dtype=object)
    expected[:] = [[(2. / 3, 1, 1, 1)], [(1, 3, 3, 1)]]
    assert_allclose(expected[0], y_pred_combined[0])
    assert_allclose(expected[1], y_pred_combined[1])

    # corner case: no overlap in folds
    pred1 = Predictions(
        y_pred=[[(1., 1, 1, 1)], None, None])
    pred2 = Predictions(
        y_pred=[None, [(1., 3, 3, 1)], None])
    pred3 = Predictions(
        y_pred=[None, None, [(1., 3, 3, 1)]])

    y_pred_combined = Predictions.combine([pred1, pred2, pred3]).y_pred

    expected = [[(1., 1, 1, 1)], [(1., 3, 3, 1)], [(1., 3, 3, 1)]]
    assert len(expected) == len(y_pred_combined)
    [assert_array_equal(x, y) for x, y in zip(y_pred_combined, expected)]


def test_combine_zero_conf():

    pred1 = Predictions(
        y_pred=[[(0, 1, 1, 1)], [(0, 1, 1, 1)]])
    pred2 = Predictions(
        y_pred=[[(1, 1.1, 1.1, 1)], [(0, 1, 1, 1)]])

    y_pred_combined = Predictions.combine([pred1, pred2]).y_pred

    expected = [[(0.5, 1.1, 1.1, 1)], [(0., 1, 1, 1)]]
    [assert_array_equal(x, y) for x, y in zip(y_pred_combined, expected)]



import numpy as np
from .feature_extractor_classifier import FeatureExtractorClassifier
from .feature_extractor_regressor import FeatureExtractorRegressor


class DrugSpectra(object):
    def __init__(self, workflow_element_names=[
            'feature_extractor_clf', 'classifier',
            'feature_extractor_reg', 'regressor']):
        self.element_names = workflow_element_names
        self.feature_extractor_classifier_workflow =\
            FeatureExtractorClassifier(self.element_names[:2])
        self.feature_extractor_regressor_workflow =\
            FeatureExtractorRegressor(self.element_names[2:])

    def train_submission(self, module_path, X_df, y_array, train_is=None):
        if train_is is None:
            train_is = slice(None, None, None)
        # Avoid setting with copy warning
        X_train_df = X_df.iloc[train_is].copy()
        y_train_array = y_array[train_is]
        y_train_clf_array = y_train_array[:, 0]
        y_train_reg_array = y_train_array[:, 1].astype(float)
        fe_clf, clf = self.feature_extractor_classifier_workflow.\
            train_submission(module_path, X_train_df, y_train_clf_array)

        # Concatenating ground truth y_proba (one-hot, derived from labels)
        # to X_train_df.
        # This makes it vulnerable to training sets that don't contain
        # all the classes. So better to use it with stratified CV.
        labels = np.unique(y_array[:, 0])
        for i, label in enumerate(labels):
            X_train_df.loc[:, label] = (y_train_clf_array == label)
        fe_reg, reg = self.feature_extractor_regressor_workflow.\
            train_submission(module_path, X_train_df, y_train_reg_array)

        # It's a bit ugly that we return the labels here, but I don't see
        # a better solution
        return labels, fe_clf, clf, fe_reg, reg

    def test_submission(self, trained_model, X_df):
        labels, fe_clf, clf, fe_reg, reg = trained_model
        y_proba_clf = self.feature_extractor_classifier_workflow.\
            test_submission((fe_clf, clf), X_df)
        # Avoid setting with copy warning
        X_df = X_df.copy()
        # Concatenating ground y_proba to X_train_df.
        for i, label in enumerate(labels):
            X_df.loc[:, label] = y_proba_clf[:, i]
        y_pred_reg = self.feature_extractor_regressor_workflow.\
            test_submission((fe_reg, reg), X_df)
        return np.concatenate([y_proba_clf, y_pred_reg.reshape(-1, 1)], axis=1)



import click

from ..hyperopt import run_hyperopt

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option('--submission', default='starting_kit', show_default=True,
              help='The kit to hyperopt. It should be located in the '
              '"submissions" folder of the starting kit.')
@click.option('--ramp-kit-dir', default='.', show_default=True,
              help='Root directory of the ramp-kit to hyperopt.')
@click.option('--ramp-data-dir', default='.', show_default=True,
              help='Directory containing the data. This directory should '
              'contain a "data" folder.')
@click.option('--ramp-submission-dir', default='submissions',
              show_default=True,
              help='Directory where the submissions are stored. It is the '
              'directory (typically called "submissions" in the ramp-kit) '
              'that contains the individual submission subdirectories.')
@click.option('--engine', default='random', show_default=True,
              help='The name of the hyperopt engine, e.g., "random".')
@click.option('--n-iter', default=10, show_default=True,
              help='The number of hyperopt iterations, inputted to the '
              'engine. The granularity is per cv fold, so if you want to '
              'fully test 7 hyperparameter combinations for example with the '
              'random engine and you have 8 CV folds, you should enter '
              '--n-iter 56')
@click.option('--save-best', is_flag=True, default=True,
              show_default=True,
              help='Specify this flag to create a <submission>_hyperopt '
              'in hte "submissions" dir with the best submission.')
def main(submission, ramp_kit_dir, ramp_data_dir, ramp_submission_dir,
         engine, n_iter, save_best):
    """Hyperopt a submission."""
    run_hyperopt(
        ramp_kit_dir=ramp_kit_dir, ramp_data_dir=ramp_data_dir,
        ramp_submission_dir=ramp_submission_dir, submission=submission,
        engine_name=engine, n_iter=n_iter, save_best=save_best)


def start():
    main()


if __name__ == '__main__':
    start()



from __future__ import division

import numpy as np
from scipy.optimize import linear_sum_assignment

from .iou import cc_iou


def _select_minipatch_tuples(y_list, minipatch):
    """
    Mask over a list selecting the tuples that lie in the minipatch

    Parameters
    ----------
    y_list : list of (row, col, radius)
        List of true or predicted objects
    minipatch : [row_min, row_max, col_min, col_max]
        Bounds of the internal scoring patch

    Returns
    -------
    y_list_cut : list of bool
        List of booleans corresponding to whether the circle is in
        the minipatch or not

    """
    if len(y_list) == 0:
        return np.array([], dtype=bool)

    row_min, row_max, col_min, col_max = minipatch

    rows, cols, _ = np.asarray(y_list).T

    y_list_cut = ((rows >= row_min) &
                  (rows < row_max) &
                  (cols >= col_min) &
                  (cols < col_max))

    return y_list_cut


def _filter_minipatch_tuples(y_list, minipatch):
    """
    Mask over a list selecting the tuples that lie in the minipatch

    Parameters
    ----------
    y_list : list of (row, col, radius)
        List of true or predicted objects
    minipatch : [row_min, row_max, col_min, col_max]
        Bounds of the internal scoring patch

    Returns
    -------
    y_list_filtered : list of (row, col, radius)
        List of filtered tuples corresponding to the scoring region

    """
    in_minipatch = _select_minipatch_tuples(y_list, minipatch)

    return [tupl for (tupl, cond) in zip(y_list, in_minipatch) if cond]


def _match_tuples(y_true, y_pred):
    """
    Given set of true and predicted (x, y, r) tuples.

    Determine the best possible match.

    Parameters
    ----------
    y_true, y_pred : list of tuples

    Returns
    -------
    (idxs_true, idxs_pred, ious)
        idxs_true, idxs_pred : indices into y_true and y_pred of matches
        ious : corresponding IOU value of each match

        The length of the 3 arrays is identical and the minimum of the length
        of y_true and y_pred
    """
    n_true = len(y_true)
    n_pred = len(y_pred)

    iou_matrix = np.empty((n_true, n_pred))

    for i in range(n_true):
        for j in range(n_pred):
            iou_matrix[i, j] = cc_iou(y_true[i], y_pred[j])

    idxs_true, idxs_pred = linear_sum_assignment(1 - iou_matrix)

    if (not idxs_true.size) or (not idxs_pred.size):
        ious = np.array([])
    else:
        ious = iou_matrix[idxs_true, idxs_pred]
    return idxs_true, idxs_pred, ious


def _count_matches(y_true, y_pred, matches, iou_threshold=0.5):
    """
    Count the number of matches.

    Parameters
    ----------
    y_true, y_pred : list of list of tuples
    matches : optional, output of _match_tuples
    iou_threshold : float

    Returns
    -------
    (n_true, n_pred_all, n_pred_correct):
        Number of true craters
        Number of total predicted craters
        Number of correctly predicted craters

    """
    val_numbers = []

    for y_true_p, y_pred_p, match_p in zip(y_true, y_pred, matches):
        n_true = len(y_true_p)
        n_pred = len(y_pred_p)

        _, _, ious = match_p
        p = (ious > iou_threshold).sum()

        val_numbers.append((n_true, n_pred, p))

    n_true, n_pred_all, n_pred_correct = np.array(val_numbers).sum(axis=0)

    return n_true, n_pred_all, n_pred_correct


def _locate_matches(y_true, y_pred, matches, iou_threshold=0.5):
    """
    Given list of list of matching craters.

    Return contiguous array of all
    craters x, y, r.

    Parameters
    ----------
    y_true, y_pred : list of list of tuples
    matches : optional, output of _match_tuples
    iou_threshold : float

    Returns
    -------
    loc_true, loc_pred
        Each is 2D array (n_patches, 3) with x, y, r columns

    """
    loc_true = []
    loc_pred = []

    for y_true_p, y_pred_p, matches_p in zip(y_true, y_pred, matches):

        for idx_true, idx_pred, iou_val in zip(*matches_p):
            if iou_val >= iou_threshold:
                loc_true.append(y_true_p[idx_true])
                loc_pred.append(y_pred_p[idx_pred])

    if loc_true:
        return np.array(loc_true), np.array(loc_pred)
    else:
        return np.empty((0, 3)), np.empty((0, 3))


def _filter_y_pred(y_pred, conf_threshold):
    """
    Given a list of list of predicted craters return those
    with a confidence value above given threshold

    Parameters
    ----------
    y_pred : list of list of tuples
    conf_threshold : float

    Returns
    -------
    y_pred_filtered : list of list of tuples

    """
    return [[detected_object[1:]
             for detected_object in y_pred_patch
             if detected_object[0] > conf_threshold]
            for y_pred_patch in y_pred]


def mask_detection_curve(y_true, y_pred, conf_thresholds):
    """
    Calculate mask detection score for different confidence thresholds.

    Parameters
    ----------
    y_true : list of list of tuples
        Tuples are of form (x, y, radius).
    y_pred : list of list of tuples
        Tuples are of form (confidence, x, y, radius).
    conf_thresholds : array-like
        The confidence threshold for which to calculate the
        precision and recall.

    Returns
    -------
    ms : array with score for each confidence threshold

    """
    from .scp import scp_single

    ms = []

    for conf_threshold in conf_thresholds:
        y_pred_above_confidence = _filter_y_pred(y_pred, conf_threshold)
        ms.append(scp_single(y_true, y_pred_above_confidence))

    return np.array(ms)


def ospa_curve(y_true, y_pred, conf_thresholds):
    """
    Calculate OSPA score for different confidence thresholds.

    Parameters
    ----------
    y_true : list of list of tuples
        Tuples are of form (x, y, radius).
    y_pred : list of list of tuples
        Tuples are of form (confidence, x, y, radius).
    conf_thresholds : array-like
        The confidence threshold for which to calculate the
        precision and recall.

    Returns
    -------
    os : array with OSPA score for each confidence threshold

    """
    from .ospa import ospa

    os = []

    for conf_threshold in conf_thresholds:
        y_pred_above_confidence = _filter_y_pred(y_pred, conf_threshold)
        os.append(ospa(y_true, y_pred_above_confidence))

    return np.array(os)



# -*- coding: utf-8 -*-
"""The :mod:`rampwf.utils.command_line` submodule provide command line
utils.
"""
from __future__ import print_function
from __future__ import unicode_literals
import re
import sys
import os
from os import listdir
from os.path import join, isdir
from collections import defaultdict
import numpy as np
import pandas as pd

from .testing import (
    assert_submission, assert_notebook, convert_notebook, blend_submissions)


def create_ramp_test_submission_parser():
    import argparse
    parser = argparse.ArgumentParser(
        prog='ramp_test_submission',
        description='Test your ramp-kit before attempting a submission.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--ramp_kit_dir',
                        default='.',
                        type=str,
                        help='Root directory of the ramp-kit to test.')
    parser.add_argument('--ramp_data_dir',
                        default='.',
                        type=str,
                        help='Directory containing the data. This directory'
                        ' should contain a "data" folder.')
    parser.add_argument('--ramp_submission_dir',
                        default='submissions',
                        type=str,
                        help='Directory where the submissions are stored. It '
                        'should contain a "submissions" directory.')
    parser.add_argument('--submission',
                        default='starting_kit',
                        type=str,
                        help='The kit to test. It should be located in the'
                        ' "submissions" folder of the starting kit. If "ALL",'
                        ' all submissions in the directory will be tested.')
    parser.add_argument('--quick-test', dest='quick_test', action='store_true',
                        help='Specify this flag to test the submission on a '
                             'small subset of the data.'
                        )
    parser.add_argument('--pickle', dest='pickle', action='store_true',
                        help='Specify this flag to pickle the submission '
                             'after training.')
    parser.add_argument('--save-output', dest='save_output',
                        action='store_true',
                        help='Specify this flag to save predictions, scores, '
                             'eventual error trace, and state after training.')
    parser.add_argument('--save-y-preds', dest='save_output',
                        action='store_true',
                        help='Specify this flag to save predictions, scores, '
                             'eventual error trace, and state after training. '
                             'Deprecated.')
    parser.add_argument('--retrain', dest='retrain',
                        action='store_true',
                        help='Specify this flag to retrain the submission '
                             'on the full training set after the CV loop.')
    return parser


def ramp_test_submission():
    parser = create_ramp_test_submission_parser()
    args = parser.parse_args()

    if args.quick_test:
        import os
        os.environ['RAMP_TEST_MODE'] = '1'

    is_pickle = False
    if args.pickle:
        is_pickle = True

    save_output = False
    if args.save_output:
        save_output = True

    retrain = False
    if args.retrain:
        retrain = True

    if args.submission == "ALL":
        ramp_submission_dir = join(args.ramp_kit_dir, 'submissions')
        submission = [directory
                      for directory in listdir(ramp_submission_dir)
                      if isdir(join(ramp_submission_dir, directory))]
    else:
        submission = [args.submission]

    for sub in submission:
        assert_submission(ramp_kit_dir=args.ramp_kit_dir,
                          ramp_data_dir=args.ramp_data_dir,
                          ramp_submission_dir=args.ramp_submission_dir,
                          submission=sub,
                          is_pickle=is_pickle,
                          save_output=save_output,
                          retrain=retrain)


def create_ramp_test_notebook_parser():
    import argparse
    parser = argparse.ArgumentParser(
        prog='ramp_test_notebook',
        description='Test your notebook before submitting a ramp.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--ramp_kit_dir',
                        default='.',
                        type=str,
                        help='Root directory of the ramp-kit to test.')
    return parser


def ramp_test_notebook():
    parser = create_ramp_test_notebook_parser()
    args = parser.parse_args()
    assert_notebook(ramp_kit_dir=args.ramp_kit_dir)


def ramp_convert_notebook():
    parser = create_ramp_test_notebook_parser()
    args = parser.parse_args()
    convert_notebook(ramp_kit_dir=args.ramp_kit_dir)


def create_ramp_blend_submissions_parser():
    import argparse
    parser = argparse.ArgumentParser(
        prog='ramp_blend_submissions',
        description='Blend several submissions.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--ramp_kit_dir',
                        default='.',
                        type=str,
                        help='Root directory of the ramp-kit to test.')
    parser.add_argument('--ramp_data_dir',
                        default='.',
                        type=str,
                        help='Directory containing the data. This directory'
                        ' should contain a "data" folder.')
    parser.add_argument('--ramp_submission_dir',
                        default='submissions',
                        type=str,
                        help='Directory where the submissions are stored. It '
                        'should contain a "submissions" directory.')
    parser.add_argument('--submissions',
                        default='ALL',
                        type=str,
                        help='The submissions to blend. They should be located'
                        ' in the "submissions" folder of the starting kit.'
                        ' Specify submissions separated by a comma without'
                        ' spaces. If "ALL", all submissions in the directory'
                        ' will be blended.')
    parser.add_argument('--save-output', dest='save_output',
                        action='store_true',
                        help='Specify this flag to save predictions '
                             'after blending.')
    parser.add_argument('--min-improvement', dest='min_improvement',
                        default='0.0',
                        help='The minimum score improvement when adding.'
                        ' submissions to the ensemble.')
    return parser


def ramp_blend_submissions():
    parser = create_ramp_blend_submissions_parser()
    args = parser.parse_args()

    save_output = False
    if args.save_output:
        save_output = True

    if args.submissions == "ALL":
        ramp_submission_dir = join(args.ramp_kit_dir, 'submissions')
        submissions = [directory
                       for directory in listdir(ramp_submission_dir)
                       if isdir(join(ramp_submission_dir, directory))]
    else:
        submissions = args.submissions.split(',')

    blend_submissions(ramp_kit_dir=args.ramp_kit_dir,
                      ramp_data_dir=args.ramp_data_dir,
                      ramp_submission_dir=args.ramp_submission_dir,
                      submissions=submissions,
                      save_output=save_output,
                      min_improvement=float(args.min_improvement))


def create_ramp_leaderboard_parser():
    import argparse
    parser = argparse.ArgumentParser(
        prog='ramp_leaderboard',
        description='RAMP leaderboard, a simple command line to display'
                    'the leaderboard.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--ramp_kit_dir',
        default='.',
        type=str,
        help='Root directory of the ramp-kit to test.'
    )
    parser.add_argument(
        '--cols',
        default=None,
        type=str,
        help='List of columns (separated by ",") to display. By default'
             'it is "train_metric,valid_metric,test_metric" where metric is'
             ' the first metric according to alphabetical order. '
             'Use --help-cols to know what are the column names. Column names '
             'are of the form "metric_step" where metric could be e.g., "nll",'
             ' or "acc", and step could be "train", "valid", or "test".'
    )
    parser.add_argument(
        '--sort_by',
        default=None,
        type=str,
        help='List of columns (separated by ",") to sort the leaderboard.'
             'By default it is "test_metric,valid_metric,train_metric"'
             'where metric is the first metric according to alphabetical '
             'order if --metric is not provided otherwise the metric '
             'provided by --metric.'
    )
    parser.add_argument(
        '--metric',
        default=None,
        type=str,
        help='Metric to display. Instead of specifying --cols, we can'
             'specify the metric to display. For example, --metric=acc is'
             'equivalent to --cols=train_acc,valid_acc,test_acc.'
    )
    parser.add_argument(
        '--precision',
        default=2,
        type=int,
        help='Precision for rounding'
    )
    parser.add_argument(
        '--asc',
        dest='asc',
        action='store_true',
        help='Sort ascending'
    )
    parser.add_argument(
        '--desc',
        dest='asc',
        action='store_false',
        help='Sort descending'
    )
    parser.add_argument(
        '--help-cols',
        dest='help_cols',
        action='store_true',
        default=False,
        help='get the list of columns'
    )
    parser.add_argument(
        '--help-metrics',
        dest='help_metrics',
        action='store_true',
        default=False,
        help='get the list of metrics'
    )
    return parser


def ramp_leaderboard():
    """
    RAMP leaderboard, a simple command line to display the leaderboard.

    IMPORTANT: order to display correctly the leaderboard
    you need to save your predictions, e.g.,
    using `ramp_test_submission --submission <name> --save-y-preds`

    :param ramp_kit_dir: folder of ramp-kit to use
    :param cols: list of columns (separated by ",") to display. By default
        it is "train_metric,valid_metric,test_metric" where metric is
        the first metric according to alphabetical order. Use --help-cols to
        know what are the column names. Column names are of the form
        "metric_step" where metric could be e.g., "nll", or "acc", and step
        could be "train", "valid", or "test".
    :param sort_by: list of columns (separated by ",") to sort the leaderboard.
        By default it is "test_metric,valid_metric,train_metric"
        where metric is the first metric according to alphabetical order if
        --metric is not provided otherwise the metric provided by --metric.
    :param asc: sort ascending if True, otherwise descending
    :param metric: metric to display. Instead of specifying --cols, we can
        specify the metric to display. For example, --metric=acc is
        equivalent to --cols=train_acc,valid_acc,test_acc.
    :param precision: precision for rounding
    :param help-cols: get the list of columns
    :param help-metrics: get the list of metrics

    Examples:

    ramp_leaderboard --metric=acc

    ramp_leaderboard --cols=train_acc,valid_acc,test_acc

    ramp_leaderboard --cols=train_nll --sort-by=train_nll,train_acc --asc
    """
    parser = create_ramp_leaderboard_parser()
    args = parser.parse_args()

    try:
        scores = _build_scores_dict(args.ramp_kit_dir)
    except (IOError, OSError) as ex:
        print(ex)
        sys.exit(1)

    if len(scores) == 0:
        print('No submissions are available.')
        print('(Please make sure that you train '
              'your submissions using `ramp_test_submission --submission '
              '<name> --save-y-preds` in order to save the predictions)')
        sys.exit(0)

    df = _build_leaderboard_df(scores, precision=args.precision)
    if args.help_cols:
        for col in df.columns:
            print(col)
        sys.exit(0)

    if args.help_metrics:
        metrics = _get_metrics(df)
        for metric in metrics:
            print(metric)
        sys.exit(0)

    if args.cols and args.metric:
        print('--cols and --metric cannot both be provided')
        sys.exit(1)

    cols = args.cols.split(',') if args.cols else None
    sort_by = args.sort_by.split(',') if args.sort_by else None

    try:
        df = _filter_and_sort_leaderboard_df(
            df,
            cols=cols,
            metric=args.metric,
            sort_by=sort_by,
            asc=args.asc
        )
    except ValueError as ex:
        print(ex)
        sys.exit(1)
    try:
        from ..externals.tabulate import tabulate
        print(tabulate(df, headers='keys', tablefmt='grid'))
    except ImportError:
        print(df)


def _filter_and_sort_leaderboard_df(
        df, cols=None, metric=None,
        sort_by=None, asc=False):
    """
    Filter and sorts rows of df.

    df is a DataFrame obtained using _build_leaderboard_df.

    Parameters
    ----------

    cols : None or list of str
        columns to take from df.
        if a column does not exist in df, the function
        will raise a ValueError.

    metric : None or str
        metric to take from df. Equivalent to
        specifying cols to ['train_metric', 'valid_metric', 'test_metric'].
        If the specified metric does not exist, the function will raise a
        ValueError.
        Note that we cannot both provide cols and metric, only one of them has
        to be provided otherwise the functions returns None.

    sort_by : None or list of str
        columns to sort by. That is, sort by the col of the first
        element sort_by, then the second element, etc.
        If one of the columns do not exist, the function
        will raise a ValuError

    asc : bool
        True if sorting ascending otherwise descending

    Returns
    -------

    pd.DataFrame : df with filtering and sorting

    Raises
    ------

    ValueError if one of the `cols` do not exist in `df`, or
    `metric` do not exist in `df`, or one of the columns in
    `sort_by` do not exist in `df`.
    """
    if cols and metric:
        return
    elif cols is not None:
        valid_cols = set(df.columns.tolist())
        for col in cols:
            if col not in valid_cols:
                cols_s = '\n'.join(list(valid_cols))
                raise ValueError(
                    'Column "{}" does not exist.'
                    ' Available columns are : \n{}'.format(col, cols_s))
        show_cols = cols
    elif metric is not None:
        if 'train_' + metric not in df.columns:
            metrics = '\n'.join(_get_metrics(df))
            raise ValueError(
                'Metric "{}" does not exist.'
                ' Available metrics are : \n{}'.format(metric, metrics))
        show_cols = [
            '{}_{}'.format(step, metric)
            for step in ('train', 'valid', 'test')]
    else:
        metrics = _get_metrics(df)
        metrics = sorted(metrics)
        metric = metrics[0]
        show_cols = [
            '{}_{}'.format(step, metric)
            for step in ('train', 'valid', 'test')]
    if sort_by is not None:
        sort_cols = sort_by
        valid_cols = set(df.columns)
        for col in sort_cols:
            if col not in valid_cols:
                cols_s = '\n'.join(list(valid_cols))
                raise ValueError(
                    'Column "{}" does not exist.'
                    ' Available columns are :\n{}'.format(col, cols_s))
    elif metric:
        sort_cols = ['{}_{}_mean'.format(step, metric)
                     for step in ('test', 'valid', 'train')]
    else:
        metrics = _get_metrics(df)
        metrics = sorted(metrics)
        metric = metrics[0]
        sort_cols = ['{}_{}'.format(step, metric)
                     for step in ('test', 'valid', 'train')]
    df = df.sort_values(by=sort_cols, ascending=asc)
    df = df[['submission'] + show_cols]
    return df


def _build_leaderboard_df(scores_dict, precision=2):
    """
    Get a pd.DataFrame representing the leaderboard.

    Each row is a submission and each column
    is the mean or std of score on train or valid or test data for a metric.
    There is a column name 'submission' for the submission names.
    The rest of column names are prefixed by the 'train' or 'valid' or 'test'
    and followed by the metric and optionally followed by '_mean' or '_std'.
    Example of column names are: 'train_acc', 'valid_acc', 'train_acc_mean',
    'train_acc_std'. 'train_acc' will be a string of the form 'mean ± std'
    whereas 'train_acc_mean' and 'train_acc_std' will be floats.

    Parameters
    ----------

    scores_dict : dict
        scores dictionary returned by _build_scores_dict

    Returns
    -------

    pd.DataFrame


    """
    rows = []
    if len(scores_dict) == 0:
        return pd.DataFrame()

    for submission_name, scores_folds in scores_dict.items():
        scs = [pd.DataFrame(s) for s in scores_folds.values()]
        # compute mean and std over folds for current submission
        mean_scores = sum([s for s in scs]) / len(scs)
        metrics = mean_scores.columns
        std_scores = np.sqrt(
            sum([s**2 for s in scs]) / len(scs) -
            mean_scores ** 2
        )
        # create row for the current submission
        row = {}
        row['submission'] = submission_name
        for step in mean_scores.index.values:
            for metric in mean_scores.columns:
                colname = '{}_{}'.format(step, metric)

                mu = mean_scores.loc[step, metric]
                fmt = '{:.' + str(precision) + 'f}'
                mu = fmt.format(mu)

                fmt = '{:.' + str(precision + 1) + 'f}'
                std = std_scores.loc[step, metric]
                std = fmt.format(std)
                row[colname] = '{} ± {}'.format(mu, std)
                row[colname + '_mean'] = mu
                row[colname + '_std'] = std
        rows.append(row)
    columns_order = [
        '{}_{}'.format(step, metric)
        for metric in metrics
        for step in ('train', 'valid', 'test')
    ]
    columns_order += [
        '{}_{}_{}'.format(step, metric, part)
        for metric in metrics
        for step in ('train', 'valid', 'test')
        for part in ('mean', 'std')
    ]
    return pd.DataFrame(rows, columns=['submission'] + columns_order)


def _get_metrics(leaderboard_df):
    """
    Get the list of metrics used in `leaderboard_df`.

    Parameters
    ----------

    leaderboard_df : pd.DataFrame
        data frame returned by _build_leaderboard_df

    Returns
    -------

    list of str

    Example of return value : ['acc', 'nll']
    """
    metrics = [
        re.search('train_(.+)_mean', col).group(1)
        for col in leaderboard_df.columns
        if col.startswith('train_') and col.endswith('_mean')
    ]
    metrics = list(set(metrics))
    return metrics


def _build_scores_dict(ramp_kit_dir='.'):
    """
    Build a nested dictionary of scores.

    Using the training_output/ folder
    of each submission.

    The structure of the folder `ramp_kit_dir` should be
    something like this:

    submissions/submission1/training_output/fold_0/scores.csv
    submissions/submission1/training_output/fold_1/scores.csv
    submissions/submission2/training_output/fold_0/scores.csv
    submissions/submission2/training_output/fold_1/scores.csv

    The scores.csv files should be like the following.
    There is one required column, "step" which specifies the step,
    it can be either 'train' or 'valid' or 'test'.
    The other columns are the metrics and are problem dependent.
    Example of a scores.csv file:

    step,acc,error,f1_70,nll
    train,0.57,0.42,0.33,1.17
    test,0.70,0.29,0.66,0.80
    valid,0.65,0.35,0.33,0.52

    The structure of the returned dict corresponding to the above
    folder structure will be like this:

    {
        'submission1': {
            0: {
                'acc': {'train': ..., 'valid': ..., 'test': ...},
                'nll': {'train': ..., 'valid': ..., 'test': ...},
            },
            1: {
                'acc': {'train': ..., 'valid': ..., 'test': ...},
                'nll': {'train': ..., 'valid': ..., 'test': ...}
            },
        },
        'submission2': {
            0: {
                'acc': {'train': ..., 'valid': ..., 'test': ...},
                'nll': {'train': ..., 'valid': ..., 'test': ...},
            },
            1: {
                'acc': {'train': ..., 'valid': ..., 'test': ...},
                'nll': {'train': ..., 'valid': ..., 'test': ...}
            },
        }
    }
    Here 0 and 1 are fold numbers.
    'acc' and 'nll' are metrics.
    """
    submissions_folder = os.path.join(ramp_kit_dir, 'submissions')
    scores = defaultdict(dict)
    for submission_name in os.listdir(submissions_folder):
        submission_folder = os.path.join(submissions_folder, submission_name)
        training_output = os.path.join(submission_folder, 'training_output')
        if not os.path.exists(training_output):
            continue
        for fold_name in os.listdir(training_output):
            if not fold_name.startswith('fold_'):
                continue
            _, fold_number = fold_name.split('_')
            fold_number = int(fold_number)
            scores_file = os.path.join(
                training_output, fold_name, 'scores.csv')
            if not os.path.exists(scores_file):
                continue
            fold_scores = pd.read_csv(scores_file)
            fold_scores = fold_scores.set_index('step').to_dict()
            scores[submission_name][fold_number] = fold_scores

    if len(scores) == 0:
        return scores
    # check consistency (all the submissions must have the same structure)
    # -- consistency of nb of folds
    nb_folds_per_submission = [len(folds) for folds in scores.values()]
    assert len(set(nb_folds_per_submission)) == 1
    # -- consistency of metric names
    metrics_per_submission = [
        frozenset(metrics.keys())
        for folds in scores.values()
        for metrics in folds.values()
    ]
    assert len(set(metrics_per_submission)) == 1

    # check correctness of scores dict of each fold of each sumbission
    # -- we should have 3 steps per fold, 'train', 'valid', 'test'
    assert all(
        set(metric.keys()) == set(('train', 'valid', 'test'))
        for folds in scores.values()
        for metrics in folds.values()
        for metric in metrics.values()
    )
    return scores



"""
Utility to import local files from the filesystem as modules.

"""
import importlib
from .sanitize import _sanitize_input


def import_module_from_source(source, name, sanitize=False):
    """Load a module from a Python source file.

    Parameters
    ----------
    source : str
        Path to the Python source file which will be loaded as a module.
    name : str
        Name to give to the module once loaded.
    sanitize: bool, default=False
        Check for blacklisted key words in code.

    Returns
    -------
    module : Python module
        Return the Python module which has been loaded.
    """
    if sanitize:
        with open(source, 'rt') as fh:
            _sanitize_input(fh.read())

    # import the module from the source file following the instructions given
    # in the importlib doc https://docs.python.org/3/library/importlib.html.
    # we do not add the module to sys.modules to make cloudpickle work.
    # see issue #232.
    spec = importlib.util.spec_from_file_location(name, source)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



import os

from ..utils.importing import import_module_from_source


class FeatureExtractor(object):
    def __init__(self, workflow_element_names=['feature_extractor']):
        self.element_names = workflow_element_names

    def train_submission(self, module_path, X_df, y_array, train_is=None):
        if train_is is None:
            train_is = slice(None, None, None)
        feature_extractor = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        fe = feature_extractor.FeatureExtractor()
        fe.fit(X_df.iloc[train_is], y_array[train_is])
        return fe

    def test_submission(self, trained_model, X_df):
        fe = trained_model
        X_test_array = fe.transform(X_df)
        return X_test_array



from __future__ import division

import numpy as np

from rampwf.score_types.soft_accuracy import SoftAccuracy


score_matrix_1 = np.array([
    [1, 0, 0],
    [0, 1, 0],
    [0, 0, 1],
])

score_matrix_2 = np.array([
    [1, 0.5, 0],
    [0.3, 1, 0.3],
    [0, 0.5, 1],
])

y_true_proba_1 = np.array([1, 0, 0])
y_true_proba_2 = np.array([0, 1, 0])
y_true_proba_3 = np.array([0.5, 0.5, 0])

y_proba_1 = np.array([1, 0, 0])
y_proba_2 = np.array([0, 1, 0])
y_proba_3 = np.array([0, 0.1, 0])
y_proba_4 = np.array([-1, 0.1, -2])
y_proba_5 = np.array([0, 0, 0])
y_proba_6 = np.array([0.5, -1, 3])


def test_soft_accuracy():
    score_1 = SoftAccuracy(score_matrix=score_matrix_1)
    assert score_1(np.array([y_true_proba_1]), np.array([y_proba_1])) == 1
    assert score_1(np.array([y_true_proba_1]), np.array([y_proba_2])) == 0
    assert score_1(np.array([y_true_proba_1]), np.array([y_proba_3])) == 0
    assert score_1(np.array([y_true_proba_1]), np.array([y_proba_4])) == 0
    assert score_1(np.array([y_true_proba_2]), np.array([y_proba_1])) == 0
    assert score_1(np.array([y_true_proba_2]), np.array([y_proba_2])) == 1
    assert score_1(np.array([y_true_proba_2]), np.array([y_proba_3])) == 1
    assert score_1(np.array([y_true_proba_2]), np.array([y_proba_4])) == 1
    assert score_1(np.array([y_true_proba_3]), np.array([y_proba_1])) == 0.5
    assert score_1(np.array([y_true_proba_3]), np.array([y_proba_2])) == 0.5
    assert score_1(np.array([y_true_proba_1]), np.array([y_proba_5])) == 0.0
    assert score_1(np.array([y_true_proba_2]), np.array([y_proba_5])) == 0.0
    assert score_1(np.array([y_true_proba_3]), np.array([y_proba_5])) == 0
    assert score_1(np.array([y_true_proba_1]), np.array([y_proba_6])) == 1 / 3
    assert score_1(np.array([y_true_proba_2]), np.array([y_proba_6])) == 0.0
    assert score_1(np.array([y_true_proba_3]), np.array([y_proba_6])) == 1 / 6
    assert score_1(np.array([y_true_proba_3]), np.array([y_proba_3])) == 0.5
    assert score_1(np.array([y_true_proba_3]), np.array([y_proba_4])) == 0.5

    score_2 = SoftAccuracy(score_matrix=score_matrix_2)
    assert score_2(np.array([y_true_proba_1]), np.array([y_proba_1])) == 1
    assert score_2(np.array([y_true_proba_1]), np.array([y_proba_2])) == 0.5
    assert score_2(np.array([y_true_proba_1]), np.array([y_proba_3])) == 0.5
    assert score_2(np.array([y_true_proba_1]), np.array([y_proba_4])) == 0.5
    assert score_2(np.array([y_true_proba_2]), np.array([y_proba_1])) == 0.3
    assert score_2(np.array([y_true_proba_2]), np.array([y_proba_2])) == 1
    assert score_2(np.array([y_true_proba_2]), np.array([y_proba_3])) == 1
    assert score_2(np.array([y_true_proba_2]), np.array([y_proba_4])) == 1
    assert score_2(np.array([y_true_proba_3]), np.array([y_proba_1])) == 0.65
    assert score_2(np.array([y_true_proba_3]), np.array([y_proba_2])) == 0.75
    assert score_2(np.array([y_true_proba_1]), np.array([y_proba_5])) == 0.0
    assert score_2(np.array([y_true_proba_2]), np.array([y_proba_5])) == 0.0
    assert score_2(np.array([y_true_proba_3]), np.array([y_proba_5])) == 0.0
    assert score_2(np.array([y_true_proba_1]), np.array([y_proba_6])) == 1 / 3
    assert score_2(np.array([y_true_proba_2]), np.array([y_proba_6])) == 0.3
    assert score_2(np.array([y_true_proba_3]), np.array([y_proba_3])) == 0.75
    assert score_2(np.array([y_true_proba_3]), np.array([y_proba_4])) == 0.75



"""A time series cross validation class.

It implements a block cross validation. We can't simply shuffle the
observations z_t since we would lose both causality and the correlation
structure that follows natural order. To formalize the issue, let us first
define formally the predictor that we will produce in the RAMP. Let the time
series be z_1, ..., z_T and the let target to predict at time t be
y_t. The target is usually a function of the future
z_{t+1}, ..., but it can be anything else. We want to learn a function
that predicts y from the past, that is y_hat_t = f(z_1, ..., z_t) = f(Z_t),
where Z_t = (z_1, ..., z_t) is the past. Now, the sample (Z_t, y_t) is a
regular (although none iid) sample from the point of view of shuffling, so we
can train on {Z_t, y_t}_{t in I_train} and test on
(Z_t, y_t)_{t in I_test}, where I_train
and I_test are arbitrary but disjunct train and test index
sets, respectively (typically produced by sklearn's `ShuffleSplit`). Using
shuffling would nevertheless allow a second order leakage from training
points to test points that preceed them, by, e.g., aggregating the training
set and adding the aggregate back as a feature. To avoid this, we use
block-CV: on each fold, all t in I_test are larger than all
t in I_train. We also make sure that all training and test
sets contain consecutive observations, so recurrent nets and similar
predictors, which rely on this, may be trained.

The block cv can be parameterized by `cv_block_size`, `n_cv`, and `period`.
`cv_block_size` is the relative size of the validation block. If it is, e.g.,
0.3, it means that all folds have a common block which is (approximately) 0.7
times the length of the sequence. `n_cv` is the number of the folds. `period`
can be used when we want that the length of each training block is a multiple
of an integer (e.g., the number of months in a year), assuring that each block
starts at the same phase (e.g., the beginning of the year).

A classical simple validation (a single training block followed by a test
block) can be done by setting `n_cv` to 1.
"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause
import numpy as np


class TimeSeries(object):
    def __init__(self, n_cv=8, cv_block_size=0.5, period=12, unit='',
                 unit_2=None):
        self.n_cv = n_cv
        self.cv_block_size = cv_block_size
        self.period = period
        self.unit = unit
        self.unit_2 = unit_2

    def get_cv(self, X, y):
        n = len(y)
        block_size = int(
            n * self.cv_block_size / self.n_cv / self.period) * self.period
        n_common_block = n - block_size * self.n_cv
        n_validation = n - n_common_block
        if self.unit_2 is None:
            print('length of common block: {} {}s'.format(
                n_common_block, self.unit))
            print('length of validation block: {} {}s'.format(
                n_validation, self.unit))
            print('length of each cv block: {} {}s'.format(
                block_size, self.unit))
        else:
            print('length of common block: {} {}s = {} {}s'.format(
                n_common_block, self.unit, n_common_block / self.period,
                self.unit_2))
            print('length of validation block: {} {}s = {} {}s'.format(
                n_validation, self.unit, n_validation / self.period,
                self. unit_2))
            print('length of each cv block: {} {}s = {} {}s'.format(
                block_size, self.unit, block_size / self.period, self.unit_2))
        for i in range(self.n_cv):
            train_is = np.arange(0, n_common_block + i * block_size)
            test_is = np.arange(n_common_block + i * block_size, n)
            yield (train_is, test_is)



from __future__ import division

import numpy as np

from .base import DetectionBaseScoreType


class SCP(DetectionBaseScoreType):
    is_lower_the_better = True
    minimum = 0.0
    maximum = 1.0

    def __init__(self, shape, name='scp', precision=3, conf_threshold=0.5,
                 minipatch=None):
        self.shape = shape
        self.name = name
        self.precision = precision
        self.conf_threshold = conf_threshold
        self.minipatch = minipatch

    def detection_score(self, y_true, y_pred):
        """
        Score based on a matching by reprojection of craters on mask-map.

        True craters are projected positively, predicted craters negatively,
        so they can cancel out. Then the sum of the absolute value of the
        residual map is taken.

        The best score value for a perfect match is 0.
        The worst score value for a given patch is the sum of all crater
        instances in both `y_true` and `y_pred`.

        Parameters
        ----------
        y_true : list of list of tuples (x, y, radius)
            List of coordinates and radius of actual craters for set of patches
        y_pred : list of list of tuples (x, y, radius)
            List of coordinates and radius of predicted craters for set of
            patches

        Returns
        -------
        float : score for a given patch, the lower the better

        """
        scps = np.array(
            [scp_single(t, p, self.shape, self.minipatch)
             for t, p in zip(y_true, y_pred)])
        return np.sum(scps[:, 0]) / np.sum(scps[:, 1:3])


def scp_single(y_true, y_pred, shape, minipatch=None):
    """
    L1 distance between superposing bounding box cylinder or prism maps.

    True craters are projected positively, predicted craters negatively,
    so they can cancel out. Then the sum of the absolute value of the
    residual map is taken.

    The best score value for a perfect match is 0.
    The worst score value for a given patch is the sum of all crater
    instances in both `y_true` and `y_pred`.

    Parameters
    ----------
    y_true : list of tuples (x, y, radius)
        List of coordinates and radius of actual craters in a patch
    y_pred : list of tuples (x, y, radius)
        List of coordinates and radius of craters predicted in the patch
    shape : tuple of int
        Shape of the main patch
    minipatch : [row_min, row_max, col_min, col_max], optional
        Bounds of the internal scoring patch (default is None)

    Returns
    -------
    float : score for a given patch, the lower the better

    """
    map_true, map_pred = circle_maps(y_true, y_pred, shape)
    if minipatch is not None:
        map_true = map_true[
            minipatch[0]:minipatch[1], minipatch[2]:minipatch[3]]
        map_pred = map_pred[
            minipatch[0]:minipatch[1], minipatch[2]:minipatch[3]]
    # Sum all the pixels
    score = np.abs(map_true - map_pred).sum()
    n_true = map_true.sum()
    n_pred = map_pred.sum()
    return score, n_true, n_pred


def project_circle(circle, image=None, shape=None,
                   normalize=True, negative=False):
    """
    Project circles on an image.

    Parameters
    ----------
    circle : array-like
        x, y, radius
    image : array-like, optional
        image on which to project the circle
    shape : tuple of ints, optional
        shape of the image on which to project the circle
    normalize : bool, optional (default is `True`)
        normalize the total surface of the circle to unity
    negative : bool, optional (default is `False`)
        subtract the circle instead of adding it

    Returns
    -------
    array-like : image with projected circle

    """
    if image is None:
        if shape is None:
            raise ValueError("Either `image` or `shape` must be defined")
        else:
            image = np.zeros(shape)
    else:
        shape = image.shape
    x, y, radius = circle
    coords = circle_coords(x, y, radius, shape=shape)
    value = 1

    if not coords[0].size:
        # corner case where circle is outside the image
        return image

    if normalize:
        value /= coords[0].size
    if negative:
        value = -value
    image[coords] += value
    return image


def circle_maps(y_true, y_pred, shape):
    """
    Create a map to compare true and predicted craters.

    The craters (circles) are projected on the map with a coefficient
    chosen so its sum is normalized to unity.

    True and predicted craters are projected with a different sign,
    so that good predictions tend to cancel out the true craters.

    Parameters
    ----------
    y_pred, y_true : array-like of shape (3, X)
        list of circle positions (x, y, radius)
    shape : tuple of ints, optional
        shape of image

    Returns
    -------
    array-like : image with projected true and predicted circles

    """
    map_true = np.zeros(shape)
    map_pred = np.zeros(shape)

    # Add true craters positively
    for circle in y_true:
        map_true = project_circle(
            circle, map_true, shape=shape, normalize=True)

    # Add predicted craters negatively
    for circle in y_pred:
        map_pred = project_circle(
            circle, map_pred, shape=shape, normalize=True)

    return map_true, map_pred


#
# Copyright scikit-image below
#
def _ellipse_in_shape(shape, center, radii, rotation=0.):
    """Generate coordinates of points within ellipse bounded by shape.

    Parameters
    ----------
    shape :  iterable of ints
        Shape of the input image.  Must be length 2.
    center : iterable of floats
        (row, column) position of center inside the given shape.
    radii : iterable of floats
        Size of two half axes (for row and column)
    rotation : float, optional
        Rotation of the ellipse defined by the above, in radians
        in range (-PI, PI), in contra clockwise direction,
        with respect to the column-axis.
    Returns
    -------
    rows : iterable of ints
        Row coordinates representing values within the ellipse.
    cols : iterable of ints
        Corresponding column coordinates representing values within the ellipse
    """
    r_lim, c_lim = np.ogrid[0:float(shape[0]), 0:float(shape[1])]
    r_org, c_org = center
    r_rad, c_rad = radii
    rotation %= np.pi
    sin_alpha, cos_alpha = np.sin(rotation), np.cos(rotation)
    r, c = (r_lim - r_org), (c_lim - c_org)
    distances = ((r * cos_alpha + c * sin_alpha) / r_rad) ** 2 \
        + ((r * sin_alpha - c * cos_alpha) / c_rad) ** 2
    return np.nonzero(distances < 1)


def ellipse(r, c, r_radius, c_radius, shape=None, rotation=0.):
    """Generate coordinates of pixels within ellipse.

    Parameters
    ----------
    r, c : double
        Centre coordinate of ellipse.
    r_radius, c_radius : double
        Minor and major semi-axes. ``(r/r_radius)**2 + (c/c_radius)**2 = 1``.
    shape : tuple, optional
        Image shape which is used to determine the maximum extent of output
        pixel coordinates. This is useful for ellipses which exceed the
        image size.
        By default the full extent of the ellipse are used.
    rotation : float, optional (default 0.)
        Set the ellipse rotation (rotation) in range (-PI, PI)
        in contra clock wise direction, so PI/2 degree means swap ellipse axis
    Returns
    -------
    rr, cc : ndarray of int
        Pixel coordinates of ellipse.
        May be used to directly index into an array, e.g.
        ``img[rr, cc] = 1``.
    Examples
    --------
    >>> from skimage.draw import ellipse
    >>> img = np.zeros((10, 12), dtype=np.uint8)
    >>> rr, cc = ellipse(5, 6, 3, 5, rotation=np.deg2rad(30))
    >>> img[rr, cc] = 1
    >>> img
    array([[0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0],
           [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 0],
           [0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0],
           [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
           [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0],
           [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0],
           [0, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]], dtype=uint8)
    Notes
    -----
    The ellipse equation::
        ((x * cos(alpha) + y * sin(alpha)) / x_radius) ** 2 +
        ((x * sin(alpha) - y * cos(alpha)) / y_radius) ** 2 = 1
    Note that the positions of `ellipse` without specified `shape` can have
    also, negative values, as this is correct on the plane. On the other hand
    using these ellipse positions for an image afterwards may lead to appearing
    on the other side of image, because ``image[-1, -1] = image[end-1, end-1]``
    >>> rr, cc = ellipse(1, 2, 3, 6)
    >>> img = np.zeros((6, 12), dtype=np.uint8)
    >>> img[rr, cc] = 1
    >>> img
    array([[1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 1],
           [1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
           [1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1]], dtype=uint8)
    """
    center = np.array([r, c])
    radii = np.array([r_radius, c_radius])
    # allow just rotation with in range +/- 180 degree
    rotation %= np.pi

    # compute rotated radii by given rotation
    r_radius_rot = abs(r_radius * np.cos(rotation)) \
        + c_radius * np.sin(rotation)
    c_radius_rot = r_radius * np.sin(rotation) \
        + abs(c_radius * np.cos(rotation))
    # The upper_left and lower_right corners of the smallest rectangle
    # containing the ellipse.
    radii_rot = np.array([r_radius_rot, c_radius_rot])
    upper_left = np.ceil(center - radii_rot).astype(int)
    lower_right = np.floor(center + radii_rot).astype(int)

    if shape is not None:
        # Constrain upper_left and lower_right by shape boundary.
        upper_left = np.maximum(upper_left, np.array([0, 0]))
        lower_right = np.minimum(lower_right, np.array(shape[:2]) - 1)

    shifted_center = center - upper_left
    bounding_shape = lower_right - upper_left + 1

    rr, cc = _ellipse_in_shape(bounding_shape, shifted_center, radii, rotation)
    rr.flags.writeable = True
    cc.flags.writeable = True
    rr += upper_left[0]
    cc += upper_left[1]
    return rr, cc


def circle_coords(r, c, radius, shape=None):
    """Generate coordinates of pixels within circle.

    Parameters
    ----------
    r, c : double
        Centre coordinate of circle.
    radius : double
        Radius of circle.
    shape : tuple, optional
        Image shape which is used to determine the maximum extent of output
        pixel coordinates. This is useful for circles that exceed the image
        size. If None, the full extent of the circle is used.
    Returns
    -------
    rr, cc : ndarray of int
        Pixel coordinates of circle.
        May be used to directly index into an array, e.g.
        ``img[rr, cc] = 1``.
    Examples
    --------
    >>> from skimage.draw import circle
    >>> img = np.zeros((10, 10), dtype=np.uint8)
    >>> rr, cc = circle(4, 4, 5)
    >>> img[rr, cc] = 1
    >>> img
    array([[0, 0, 1, 1, 1, 1, 1, 0, 0, 0],
           [0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
           [1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
           [1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
           [1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
           [1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
           [1, 1, 1, 1, 1, 1, 1, 1, 1, 0],
           [0, 1, 1, 1, 1, 1, 1, 1, 0, 0],
           [0, 0, 1, 1, 1, 1, 1, 0, 0, 0],
           [0, 0, 0, 0, 0, 0, 0, 0, 0, 0]], dtype=uint8)
    """
    return ellipse(r, c, radius, radius, shape)



"""A clustering (transfer learning) cross validation class.

Training and test data consist of a set of events ad each event has an
unknown number of clusters. At training time we know the classes (clusters) of
each event. At test time we only know which instances belong to the same event.
Events share statistical properties which can be exploited.

From another point of view, the setup is transfer learning: each event is
a new task, and classed in the task share statistical properties.

We use sklearn `ShuffleSplit` over the _events_, so train/test cut does not
cut into events (each event or task is either completely in the training or
completely in the test fold).
"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause
import numpy as np
from sklearn.model_selection import ShuffleSplit


class Clustering(object):
    def __init__(self, n_cv=1, cv_test_size=0.5, random_state=57):
        self.n_cv = n_cv
        self.cv_test_size = cv_test_size
        self.random_state = random_state

    def get_cv(self, X, y):
        unique_event_ids = np.unique(y[:, 0])
        event_cv = ShuffleSplit(
            n_splits=self.n_cv, test_size=self.cv_test_size,
            random_state=self.random_state)
        for train_event_is, test_event_is in event_cv.split(unique_event_ids):
            train_is = np.where(
                np.in1d(y[:, 0], unique_event_ids[train_event_is]))[0]
            test_is = np.where(
                np.in1d(y[:, 0], unique_event_ids[test_event_is]))[0]
            yield train_is, test_is



from __future__ import division

import os

import numpy as np

from ..utils.importing import import_module_from_source


class SimplifiedImageClassifier(object):
    """
    SimplifiedImageClassifier workflow.

    This workflow is used to train image classification tasks, typically when
    the dataset cannot be stored in memory. It is a simplified version
    of the `ImageClassifier` workflow where there is no batch generator
    and no image preprocessor.
    Submissions need to contain one file, which by default by is named
    image_classifier.py (it can be modified by changing
    `workflow_element_names`).
    image_classifier.py needs an `ImageClassifier` class, which implements
    `fit` and `predict_proba`, where both `fit` and `predict_proba` take
    as input an instance of `ImageLoader`.

    Parameters
    ==========

    n_classes : int
        Total number of classes.

    """

    def __init__(self, n_classes, workflow_element_names=['image_classifier']):
        self.n_classes = n_classes
        self.element_names = workflow_element_names

    def train_submission(self, module_path, folder_X_array, y_array,
                         train_is=None):
        """Train an image classifier.

        module_path : str
            module where the submission is. the folder of the module
            have to contain image_classifier.py.
        X_array : ArrayContainer vector of int
            vector of image IDs to train on
            (it is named X_array to be coherent with the current API,
             but as said here, it does not represent the data itself,
             only image IDs).
        y_array : vector of int
            vector of image labels corresponding to X_train
        train_is : vector of int
           indices from X_array to train on
        """
        folder, X_array = folder_X_array
        if train_is is None:
            train_is = slice(None, None, None)
        image_classifier = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        clf = image_classifier.ImageClassifier()
        img_loader = ImageLoader(
            X_array[train_is], y_array[train_is],
            folder=folder,
            n_classes=self.n_classes)
        clf.fit(img_loader)
        return clf

    def test_submission(self, trained_model, folder_X_array):
        """Test an image classifier.

        trained_model : tuple (function, Classifier)
            tuple of a trained model returned by `train_submission`.
        X_array : ArrayContainer of int
            vector of image IDs to test on.
            (it is named X_array to be coherent with the current API,
             but as said here, it does not represent the data itself,
             only image IDs).
        """
        folder, X_array = folder_X_array
        clf = trained_model
        test_img_loader = ImageLoader(
            X_array, None,
            folder=folder,
            n_classes=self.n_classes
        )
        y_proba = clf.predict_proba(test_img_loader)
        return y_proba


def _image_transform(x, transforms):
    from skimage.transform import rotate
    for t in transforms:
        if t['name'] == 'rotate':
            angle = np.random.random() * (
                t['u_angle'] - t['l_angle']) + t['l_angle']
            rotate(x, angle, preserve_range=True)
    return x


class ImageLoader(object):
    """
    Load and image and optionally its label.

    In image_classifier.py, both `fit` and `predict_proba` take as input
    an instance of `ImageLoader`.
    ImageLoader is used in `fit` and `predict_proba` to either load one image
    and its corresponding label  (at training time), or one image (at test
    time).
    Images are loaded by using the method `load`.

    Parameters
    ==========

    X_array : ArrayContainer of int
        vector of image IDs to train on
         (it is named X_array to be coherent with the current API,
         but as said here, it does not represent the data itself,
         only image IDs).

    y_array : vector of int or None
        vector of image labels corresponding to `X_array`.
        At test time, it is `None`.

    folder : str
        folder where the images are

    n_classes : int
        Total number of classes.
    """

    def __init__(self, X_array, y_array, folder, n_classes):
        self.X_array = X_array
        self.y_array = y_array
        self.folder = folder
        self.n_classes = n_classes
        self.nb_examples = len(X_array)

    def load(self, index):
        """
        Load and image and optionally its label.

        Load one image and its corresponding label (at training time),
        or one image (at test time).

        Parameters
        ==========

        index : int
            Index of the image to load.
            It should in between 0 and self.nb_examples - 1

        Returns
        =======

        either a tuple `(x, y)` or `x`, where:
            - x is a numpy array of shape (height, width, nb_color_channels),
              and corresponds to the image of the requested `index`.
            - y is an integer, corresponding to the class of `x`.
        At training time, `y_array` is given, and `load` returns
        a tuple (x, y).
        At test time, `y_array` is `None`, and `load` returns `x`.
        """
        from skimage.io import imread

        if index < 0 or index >= self.nb_examples:
            raise IndexError("list index out of range")

        x = self.X_array[index]
        filename = os.path.join(self.folder, '{}'.format(x))
        x = imread(filename)
        if self.y_array is not None:
            y = self.y_array[index]
            return x, y
        else:
            return x

    def parallel_load(self, indexes, transforms=None):
        """
        Load and image and optionally its label.

        Load one image and its corresponding label (at training time),
        or one image (at test time).

        Parameters
        ==========

        index : int
            Index of the image to load.
            It should in between 0 and self.nb_examples - 1

        Returns
        =======

        either a tuple `(x, y)` or `x`, where:
            - x is a numpy array of shape (height, width, nb_color_channels),
              and corresponds to the image of the requested `index`.
            - y is an integer, corresponding to the class of `x`.
        At training time, `y_array` is given, and `load` returns
        a tuple (x, y).
        At test time, `y_array` is `None`, and `load` returns `x`.
        """
        from skimage.io import imread
        from joblib import delayed, Parallel, cpu_count

        for index in indexes:
            assert 0 <= index < self.nb_examples

        n_jobs = cpu_count()
        filenames = [
            os.path.join(self.folder, '{}'.format(self.X_array[index]))
            for index in indexes]
        xs = Parallel(n_jobs=n_jobs, backend='threading')(
            delayed(imread)(filename) for filename in filenames)

        if transforms is not None:
            from functools import partial
            transform = partial(_image_transform, transforms=transforms)
            xs = Parallel(n_jobs=n_jobs, backend='threading')(
                delayed(transform)(x) for x in xs)

        if self.y_array is not None:
            ys = [self.y_array[index] for index in indexes]
            return xs, ys
        else:
            return xs

    def __iter__(self):
        for i in range(self.nb_examples):
            yield self.load(i)

    def __len__(self):
        return self.nb_examples



from __future__ import division

import numpy as np
from sklearn.utils import indices_to_mask

from .base import DetectionBaseScoreType
from .util import _select_minipatch_tuples, _match_tuples


class OSPA(DetectionBaseScoreType):
    """
    Optimal Subpattern Assignment (OSPA) metric for IoU score.

    This metric provides a coherent way to compute the miss-distance
    between the detection and alignment of objects. Among all
    combinations of true/predicted pairs, if finds the best alignment
    to minimise the distance, and still takes into account missing
    or in-excess predicted values through a cardinality score.

    The lower the value the smaller the distance.

    Arguments
    ---------
    name : str, optional
        Method name
    precision : int, optional
        Rounding precision for the score (default is 2)
    conf_threshold : float, optional
        Confidence threshold value use for the Average Precision
        measurement (default is 0.5)
    minipatch : [row_min, row_max, col_min, col_max], optional
        Bounds of the internal scoring patch (default is None)

    References
    ----------
    http://www.dominic.schuhmacher.name/papers/ospa.pdf

    """
    is_lower_the_better = True
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='ospa', precision=3, conf_threshold=0.5,
                 minipatch=None):
        self.name = name
        self.precision = precision
        self.conf_threshold = conf_threshold
        self.minipatch = minipatch

    def detection_score(self, y_true, y_pred):
        """
        Compute the OSPA score

        Parameters
        ----------
        y_true, y_pred : list of list of tuples

        Returns
        -------
        float: distance between input arrays

        """
        ospa_score = ospa(y_true, y_pred, self.minipatch)
        return 1 - ospa_score


def ospa(y_true, y_pred, minipatch=None):
    """
    Optimal Subpattern Assignment (OSPA) metric for IoU score.

    This metric provides a coherent way to compute the miss-distance
    between the detection and alignment of objects. Among all
    combinations of true/predicted pairs, if finds the best alignment
    to minimise the distance, and still takes into account missing
    or in-excess predicted values through a cardinality score.

    The lower the value the smaller the distance.

    Parameters
    ----------
    y_true, y_pred : list of list of tuples
    minipatch : [row_min, row_max, col_min, col_max], optional
        Bounds of the internal scoring patch (default is None)

    Returns
    -------
    float: distance between input arrays

    References
    ----------
    http://www.dominic.schuhmacher.name/papers/ospa.pdf

    """
    ospas = np.array([ospa_single(t, p, minipatch)
                      for t, p in zip(y_true, y_pred)])

    ious, n_count, n_total = ospas.sum(axis=0)

    if n_total == 0:
        return 0

    return ious / n_total


def ospa_single(y_true, y_pred, minipatch=None):
    """
    OSPA score on single patch. See docstring of `ospa` for more info.

    Parameters
    ----------
    y_true, y_pred : ndarray of shape (3, x)
        arrays of (x, y, radius)
    minipatch : [row_min, row_max, col_min, col_max], optional
        Bounds of the internal scoring region (default is None)

    Returns
    -------
    (iou_sum, n_pred, n_total)
        float - sum of ious of matched entries
        int - number of matched entries
        int - total number of entries

    """
    n_true = len(y_true)
    n_pred = len(y_pred)

    # No craters and none found
    if n_true == 0 and n_pred == 0:
        return 0, 0, 0

    # Mask of entries that lie within the minipatch
    if minipatch is not None:
        true_in_minipatch = _select_minipatch_tuples(y_true, minipatch)
        pred_in_minipatch = _select_minipatch_tuples(y_pred, minipatch)
    else:
        true_in_minipatch = np.ones(len(y_true)).astype(bool)
        pred_in_minipatch = np.ones(len(y_pred)).astype(bool)

    n_minipatch = true_in_minipatch.sum() + pred_in_minipatch.sum()

    # No craters and some found or existing craters but non found
    if n_true == 0 or n_pred == 0:
        return 0, 0, n_minipatch

    # First matching
    id_true, id_pred, ious = _match_tuples(y_true, y_pred)

    # For each set of entries (true and pred) create an array with
    # the iou corresponding to each object
    iou_true = np.zeros(len(y_true))
    iou_true[id_true] = ious
    iou_pred = np.zeros(len(y_pred))
    iou_pred[id_pred] = ious

    # Mask of matched entries
    true_matched = indices_to_mask(id_true, n_true)
    pred_matched = indices_to_mask(id_pred, n_pred)

    # Counting
    true_count = true_matched & true_in_minipatch
    pred_count = pred_matched & pred_in_minipatch

    # IoU computation on the final list
    iou_global = iou_true[true_count].sum() + iou_pred[pred_count].sum()
    n_count = true_count.sum() + pred_count.sum()

    return iou_global, n_count, n_minipatch



"""A time series feature extractor followed by a regressor.

Train and test a time series feature extractor followed by a regressor.

The input object is an `xarray` `Dataset`, containing possibly several
`DataArrays` corresponding to the input sequence. It contains a special burn
in period in the beginning (carried by X_ds.n_burn_in) for which we do not
give ground truth and we do not require the user to provide predictions.
The ground truth sequence `y_array` in train and the output of the user
submission `ts_fe.transform` are thus `n_burn_in` shorter than the input
sequence `X_ds`, making the training and testing slightly complicated.
"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause
import numpy as np
from .ts_feature_extractor import TimeSeriesFeatureExtractor
from .regressor import Regressor


class ElNino(object):
    def __init__(self, check_sizes, check_indexs, workflow_element_names=[
            'ts_feature_extractor', 'regressor']):
        self.element_names = workflow_element_names
        self.ts_feature_extractor_workflow = TimeSeriesFeatureExtractor(
            check_sizes, check_indexs, [self.element_names[0]])
        self.regressor_workflow = Regressor([self.element_names[1]])

    def train_submission(self, module_path, X_ds, y_array, train_is=None):
        """
        Train a time series feature extractor + regressor workflow.

        `X_ds` is `n_burn_in` longer than `y_array` since `y_array` contains
        targets without the initial burn in period. `train_is` are wrt
        `y_array`, so `X_ds` has to be _extended_ by `n_burn_in` when sent to
        the time series feature extractor.

        """
        if train_is is None:
            # slice doesn't work here because of the way `extended_train_is`
            # is computed below
            train_is = np.arange(len(y_array))
        ts_fe = self.ts_feature_extractor_workflow.train_submission(
            module_path, X_ds, y_array)

        n_burn_in = X_ds.n_burn_in
        # X_ds contains burn-in so it needs to be extended by n_burn_in
        # timesteps. This assumes that train_is is a block of consecutive
        # time points.
        burn_in_range = np.arange(train_is[-1], train_is[-1] + n_burn_in)
        extended_train_is = np.concatenate((train_is, burn_in_range))
        X_train_ds = X_ds.isel(time=extended_train_is)
        # At this point X_train_ds is n_burn_in longer than y_array[train_is]

        # ts_fe.transform should return an array corresponding to time points
        # without burn in, so X_train_array and y_array[train_is] should now
        # have the same length.
        X_train_array = self.ts_feature_extractor_workflow.test_submission(
            ts_fe, X_train_ds)

        reg = self.regressor_workflow.train_submission(
            module_path, X_train_array, y_array[train_is])
        return ts_fe, reg

    def test_submission(self, trained_model, X_ds):
        ts_fe, reg = trained_model
        X_test_array = self.ts_feature_extractor_workflow.test_submission(
            ts_fe, X_ds)
        y_pred = self.regressor_workflow.test_submission(reg, X_test_array)
        return y_pred



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
style.py is a part of colored.

Copyright 2014-2017 Dimitris Zlatanidis <d.zlatanidis@gmail.com>
All rights reserved.

Colored is very simple Python library for color and formatting in terminal.

https://github.com/dslackw/colored

colored is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""


class style(object):

    ESC = "\x1b["
    END = "m"

    names = {
        'BOLD': 1,
        'DIM': 2,
        'UNDERLINED': 4,
        'BLINK': 5,
        'REVERSE': 7,
        'HIDDEN': 8,
        'RESET': 0,
        'RES_BOLD': 21,
        'RES_DIM': 22,
        'RES_UNDERLINED': 24,
        'RES_BLINK': 25,
        'RES_REVERSE': 27,
        'RES_HIDDEN': 28
    }

    for color, num in names.items():
        vars()[color] = '{}{}{}'.format(ESC, num, END)



from .base import BaseScoreType


class ClassifierBaseScoreType(BaseScoreType):
    def score_function(self, ground_truths, predictions, valid_indexes=None):
        self.label_names = ground_truths.label_names
        if valid_indexes is None:
            valid_indexes = slice(None, None, None)
        y_pred_label_index = predictions.y_pred_label_index[valid_indexes]
        y_true_label_index = ground_truths.y_pred_label_index[valid_indexes]
        self.check_y_pred_dimensions(y_true_label_index, y_pred_label_index)
        return self.__call__(y_true_label_index, y_pred_label_index)



from . import cvs
from . import externals
from . import prediction_types
from . import score_types
from . import utils
from . import workflows
from ._version import __version__


__all__ = [
    'cvs',
    'externals',
    'prediction_types',
    'score_types',
    'utils',
    'workflows',
    '__version__'
]



import os
import ast

import click
import pandas as pd

from ...externals.tabulate import tabulate

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class PythonLiteralOption(click.Option):

    def type_cast_value(self, ctx, value):
        try:
            return ast.literal_eval(value)
        except:  # noqa: E722
            raise click.BadParameter(value)


def _load_score_submission(submission_path, metric, step):
    """Load the score for a single submission."""
    training_output_path = os.path.join(submission_path, 'training_output')
    if not os.path.isdir(training_output_path):
        return None
    folds_path = [
        os.path.join(training_output_path, fold_name)
        for fold_name in os.listdir(training_output_path)
        if (os.path.isdir(os.path.join(training_output_path, fold_name)) and
            'fold_' in fold_name)
    ]
    data = {}
    for fold_id, path in enumerate(folds_path):
        score_path = os.path.join(path, 'scores.csv')
        if not os.path.exists(score_path):
            return
        scores = pd.read_csv(score_path, index_col=0)
        scores.columns.name = 'score'
        data[fold_id] = scores
    df = pd.concat(data, names=['fold'])
    metric = metric if metric else slice(None)
    step = step if step else slice(None)
    return df.loc[(slice(None), step), metric]


def _bagged_table_and_headers(all_submissions):
    subs = []
    valid_scores = []
    test_scores = []
    for sub, path in all_submissions.items():
        bagged_scores_path = os.path.join(
            path, 'training_output', 'bagged_scores.csv')
        if not os.path.isfile(bagged_scores_path):
            continue
        bagged_scores_df = pd.read_csv(bagged_scores_path)
        n_folds = len(bagged_scores_df) // 2
        subs.append(sub)
        valid_scores.append(bagged_scores_df.iloc[n_folds - 1, 2])
        test_scores.append(bagged_scores_df.iloc[2 * n_folds - 1, 2])
        metric = bagged_scores_df.columns[2]
    df = pd.DataFrame()
    df['submission'] = subs
    df['valid {}'.format(metric)] = valid_scores
    df['test {}'.format(metric)] = test_scores
    headers = df.columns.to_numpy()
    return df, headers


def _mean_table_and_headers(all_submissions, metric, step):
    data = {}
    for sub_name, sub_path in all_submissions.items():
        scores = _load_score_submission(sub_path, metric, step)
        if scores is None:
            continue
        data[sub_name] = scores
    df = pd.concat(data, names=['submission'])
    df = df.unstack(level=['step'])
    df = pd.concat([df.groupby('submission').mean(),
                    df.groupby('submission').std()],
                   keys=['mean', 'std'], axis=1, names=['stat'])
    df = df.reorder_levels([1, 2, 0], axis=1)
    step = ['train', 'valid', 'test'] if not step else step
    df = (df.sort_index(axis=1, level=0)
            .reindex(labels=step, level='step', axis=1))
    headers = (["\n".join(df.columns.names)] +
               ["\n".join(col_names)
                for col_names in df.columns.to_numpy()])
    return df, headers


@click.group(context_settings=CONTEXT_SETTINGS)
def main():
    """Command-line to show information about local submissions."""
    pass


@main.command()
@click.option("--ramp-kit-dir", default='.', show_default=True,
              help='Root directory of the ramp-kit to retrieved the train '
              'submission.')
@click.option("--metric", cls=PythonLiteralOption, default="[]",
              show_default=True,
              help='A list of the metric to report')
@click.option("--step", cls=PythonLiteralOption, default="[]",
              show_default=True,
              help='A list of the processing to report. Choices are '
              '{"train" , "valid", "test"}')
@click.option("--sort-by", cls=PythonLiteralOption, default="[]",
              show_default=True,
              help='Give the metric, step, and stat to use for sorting.')
@click.option("--ascending/--descending", default=True, show_default=True,
              help='Sort in ascending or descending order.')
@click.option("--precision", default=2, show_default=True,
              help='The precision for the different metrics reported.')
@click.option("--bagged/--mean", default=True, show_default=True,
              help='Bagged or mean scores.')
def leaderboard(ramp_kit_dir, metric, step, sort_by, ascending, precision,
                bagged):
    """Display the leaderboard for all the local submissions."""
    path_submissions = os.path.join(ramp_kit_dir, 'submissions')
    all_submissions = {
        sub: os.path.join(path_submissions, sub)
        for sub in os.listdir(path_submissions)
        if os.path.isdir(os.path.join(path_submissions, sub))
    }
    if bagged:  # bagged scores
        df, headers = _bagged_table_and_headers(all_submissions)
    else:  # mean scores with std
        df, headers = _mean_table_and_headers(all_submissions, metric, step)

    df = df.round(precision)
    if sort_by:
        df = df.sort_values(sort_by, ascending=ascending, axis=0)

    click.echo(tabulate(df, headers=headers, tablefmt='grid'))


def start():
    main()


if __name__ == '__main__':
    start()



"""
Precision and recall (and location and size precision) based on a best
match (given a specific confidence threshold).

"""
from __future__ import division

import numpy as np

from .base import DetectionBaseIOUScoreType
from .util import (
    _match_tuples, _count_matches, _locate_matches, _filter_minipatch_tuples)


class DetectionPrecision(DetectionBaseIOUScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0
    _name = 'precision'

    def detection_score(self, y_true, y_pred):
        return precision(y_true, y_pred, minipatch=self.minipatch,
                         iou_threshold=self.iou_threshold)


class DetectionRecall(DetectionBaseIOUScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0
    _name = 'recall'

    def detection_score(self, y_true, y_pred):
        return recall(y_true, y_pred, minipatch=self.minipatch,
                      iou_threshold=self.iou_threshold)


class MADCenter(DetectionBaseIOUScoreType):
    is_lower_the_better = True
    minimum = 0.0
    maximum = np.inf
    _name = 'mad_center'

    def detection_score(self, y_true, y_pred):
        return mad_center(y_true, y_pred, minipatch=self.minipatch,
                          iou_threshold=self.iou_threshold)


class MADRadius(DetectionBaseIOUScoreType):
    is_lower_the_better = True
    minimum = 0.0
    maximum = np.inf
    _name = 'mad_radius'

    def detection_score(self, y_true, y_pred):
        return mad_radius(y_true, y_pred, minipatch=self.minipatch,
                          iou_threshold=self.iou_threshold)


def precision(y_true, y_pred, matches=None, iou_threshold=0.5,
              minipatch=None):
    """
    Precision score (fraction of correct predictions).

    Parameters
    ----------
    y_true, y_pred : list of list of tuples
    matches : optional, output of _match_tuples
    iou_threshold : float
        Threshold to determine match
    minipatch : [row_min, row_max, col_min, col_max], optional
        Bounds of the internal scoring patch (default is None)

    Returns
    -------
    precision_score : float [0 - 1]

    """
    # precision is calculated relative to the number of predicted objects,
    # so we filter for the scoring region based on the predictions
    if minipatch is not None:
        y_pred = [_filter_minipatch_tuples(y_patch, minipatch)
                  for y_patch in y_pred]

    if matches is None:
        matches = [_match_tuples(yp_true, yp_pred)
                   for yp_true, yp_pred in zip(y_true, y_pred)]

    n_true, n_pred_all, n_pred_correct = _count_matches(
        y_true, y_pred, matches, iou_threshold=iou_threshold)

    return n_pred_correct / n_pred_all


def recall(y_true, y_pred, matches=None, iou_threshold=0.5,
           minipatch=None):
    """
    Recall score (fraction of true objects that are predicted).

    Parameters
    ----------
    y_true, y_pred : list of list of tuples
    matches : optional, output of _match_tuples
    iou_threshold : float
        Threshold to determine match
    minipatch : [row_min, row_max, col_min, col_max], optional
        Bounds of the internal scoring patch (default is None)

    Returns
    -------
    recall_score : float [0 - 1]

    """
    # recall is calculated relative to the number of true objects,
    # so we filter for the scoring region based on the true labels
    if minipatch is not None:
        y_true = [_filter_minipatch_tuples(y_patch, minipatch)
                  for y_patch in y_true]

    if matches is None:
        matches = [_match_tuples(yp_true, yp_pred)
                   for yp_true, yp_pred in zip(y_true, y_pred)]

    n_true, n_pred_all, n_pred_correct = _count_matches(
        y_true, y_pred, matches, iou_threshold=iou_threshold)

    return n_pred_correct / n_true


def mad_radius(y_true, y_pred, matches=None, iou_threshold=0.5,
               minipatch=None):
    """
    Relative Mean absolute deviation (MAD) of the radius.

    Parameters
    ----------
    y_true, y_pred : list of list of tuples
    matches : optional, output of _match_tuples
    iou_threshold : float
        Threshold to determine match
    minipatch : [row_min, row_max, col_min, col_max], optional
        Bounds of the internal scoring patch (default is None)

    Returns
    -------
    mad_radius : float > 0
    """
    if minipatch is not None:
        y_true = [_filter_minipatch_tuples(y_patch, minipatch)
                  for y_patch in y_true]

    if matches is None:
        matches = [_match_tuples(yp_true, yp_pred)
                   for yp_true, yp_pred in zip(y_true, y_pred)]

    loc_true, loc_pred = _locate_matches(
        y_true, y_pred, matches, iou_threshold=iou_threshold)

    _, _, rad_true = loc_true.T
    _, _, rad_pred = loc_pred.T

    if len(rad_true) == 0:
        return np.nan

    return np.abs((rad_pred - rad_true) / rad_true).mean()


def mad_center(y_true, y_pred, matches=None, iou_threshold=0.5,
               minipatch=None):
    """
    Relative Mean absolute deviation (MAD) of the center.

    (relative to the radius of the true object).

    Parameters
    ----------
    y_true, y_pred : list of list of tuples
    matches : optional, output of _match_tuples
    iou_threshold : float
        Threshold to determine match
    minipatch : [row_min, row_max, col_min, col_max], optional
        Bounds of the internal scoring patch (default is None)

    Returns
    -------
    mad_center : float > 0
    """
    if minipatch is not None:
        y_true = [_filter_minipatch_tuples(y_patch, minipatch)
                  for y_patch in y_true]

    if matches is None:
        matches = [_match_tuples(yp_true, yp_pred)
                   for yp_true, yp_pred in zip(y_true, y_pred)]

    loc_true, loc_pred = _locate_matches(
        y_true, y_pred, matches, iou_threshold=iou_threshold)

    x_true, y_true, rad_true = loc_true.T
    x_pred, y_pred, _ = loc_pred.T

    if len(x_true) == 0:
        return np.nan

    d = np.sqrt((x_pred - x_true) ** 2 + (y_pred - y_true) ** 2)

    return np.abs(d / rad_true).mean()



import numpy as np


def combine_predictions(Predictions, predictions_list, index_list):
    """Combine predictions predictions_list[index_list].

    Parameters
    ----------
    Predictions : a Predictions type
        Needed to call combine.
    predictions_list : list of instances of Predictions
    index_list : list of integers
        Indices of the submissions to combine (possibly with replacement).

    Returns
    -------
    combined_predictions : instance of Predictions
    """
    predictions_list_to_combine = [predictions_list[i] for i in index_list]
    combined_predictions = Predictions.combine(predictions_list_to_combine)
    return combined_predictions


def get_score_cv_bags(score_type, predictions_list, ground_truths,
                      test_is_list=None):
    """
    Compute the bagged scores of the predictions in predictions_list.

    test_is_list (list of list of integer indexes) controls which points
    in which fold participate in the combination. We return the
    combined predictions and a list of scores, where each element i is the
    score of the combination of the first i+1 folds.

    Parameters
    ----------
    score_type : instance implementing BaseScoreType signature
    predictions_list : list of instances of Predictions
    ground_truths : instance of Predictions
    test_is_list : list of list of integers
        Indices of points that should be bagged in each prediction. If None,
        the full prediction vectors will be bagged.
    Returns
    -------
    combined_predictions : instance of Predictions
    score_cv_bags : list of scores (typically floats)
    """
    Predictions = type(ground_truths)
    if test_is_list is None:  # we combine the full list
        test_is_list = [range(len(predictions.y_pred))
                        for predictions in predictions_list]

    y_comb = np.array(
        [Predictions(n_samples=len(ground_truths.y_pred))
         for _ in predictions_list])
    score_cv_bags = []
    for i, test_is in enumerate(test_is_list):
        # setting valid fold indexes of points to be combined
        y_comb[i].set_valid_in_train(predictions_list[i], test_is)
        # combine first i folds
        combined_predictions = Predictions.combine(y_comb[:i + 1])
        # get indexes of points with at least one prediction in
        # combined_predictions
        valid_indexes = combined_predictions.valid_indexes
        # score the combined predictions
        score_cv_bags.append(score_type.score_function(
            ground_truths, combined_predictions, valid_indexes))
        # Alex' old suggestion: maybe use masked arrays rather than passing
        # valid_indexes
    # TODO: will crash if len(test_is_list) == 0
    return combined_predictions, score_cv_bags


def _get_next_best_submission(predictions_list, ground_truths,
                              score_type, best_index_list,
                              min_improvement=0.0):
    """Find net best submission if added to predictions_list[best_index_list].

    Find the model that minimizes the score if added to
    predictions_list[best_index_list] using score_type.score_function.
    If there is no model improving the input
    combination, the input best_index_list is returned. Otherwise the index of
    the best model is added to the list. We could also return the combined
    prediction (for efficiency, so the combination would not have to be done
    each time; right now the algo is quadratic), but I don't think any
    meaningful rule will be associative, in which case we should redo the
    combination from scratch each time the set changes. Since mostly
    combination = mean, we could maintain the sum and the number of models, but
    it would be a bit bulky. We'll see how this evolves.

    Parameters
    ----------
    predictions_list : list of instances of Predictions
        Each element of the list is an instance of Predictions of a model
        on the same (cross-validation valid) data points.
    score_type : instance implementing BaseScoreType signature
        The score to improve by adding one submission to the ensemble.
    ground_truths : instance of Predictions
        The ground truth.
    best_index_list : list of integers
        Indices of the current best model.

    Returns
    -------
    best_index_list : list of integers
        Indices of the models in the new combination. If the same as input,
        no models were found improving the score.
    """
    Predictions = type(ground_truths)
    best_predictions = combine_predictions(
        Predictions, predictions_list, best_index_list)
    best_score = score_type.score_function(ground_truths, best_predictions)
    best_index = -1
    # Combination with replacement, what Caruana suggests. Basically, if a
    # model is added several times, it's upweighted, leading to
    # integer-weighted ensembles.
    r = np.arange(len(predictions_list))
    # Randomization doesn't matter, only in case of exact equality.
    # np.random.shuffle(r)
    # print r
    for i in r:
        # try to append the ith prediction to the current best predictions
        new_index_list = np.append(best_index_list, i)
        combined_predictions = combine_predictions(
            Predictions, predictions_list, new_index_list)
        new_score = score_type.score_function(
            ground_truths, combined_predictions)
        if score_type.is_lower_the_better:
            is_improved = new_score < best_score - min_improvement
        else:
            is_improved = new_score > best_score + min_improvement
        if is_improved:
            best_predictions = combined_predictions
            best_index = i
            best_score = new_score
    if best_index > -1:
        return np.append(best_index_list, best_index), best_score
    else:
        return best_index_list, best_score


def blend_on_fold(predictions_list, ground_truths_valid, score_type,
                  max_n_ensemble=80, min_improvement=0.0):
    """Construct the best model combination on a single fold.

    Using greedy forward selection with replacement. See
    http://www.cs.cornell.edu/~caruana/ctp/ct.papers/
    caruana.icml04.icdm06long.pdf.
    Then sets foldwise contributivity.

    Parameters
    ----------
    force_ensemble : boolean
        To force include deleted models
    """
    # The submissions must have is_to_ensemble set to True. It is for
    # fogetting models. Users can also delete models in which case
    # we make is_valid false. We then only use these models if
    # force_ensemble is True.
    # We can further bag here which should be handled in config (or
    # ramp table.) Or we could bag in get_next_best_single_fold
    if len(predictions_list) == 0:
        return None, None, None, None
    valid_scores = [score_type.score_function(ground_truths_valid, predictions)
                    for predictions in predictions_list]
    if score_type.is_lower_the_better:
        best_prediction_index = np.argmin(valid_scores)
    else:
        best_prediction_index = np.argmax(valid_scores)
    score = valid_scores[best_prediction_index]
    best_index_list = np.array([best_prediction_index])
    is_improved = True
    while is_improved and len(best_index_list) < max_n_ensemble:
        print('\t{}: {}'.format(best_index_list, score))
        old_best_index_list = best_index_list
        best_index_list, score = _get_next_best_submission(
            predictions_list, ground_truths_valid, score_type, best_index_list,
            min_improvement)
        is_improved = len(best_index_list) != len(old_best_index_list)
    return best_index_list
    # we share a unit of 1. among the contributive submissions
    # unit_contributivity = 1. / len(best_index_list)
    # for i in best_index_list:
    #     selected_submissions_on_fold[i].contributivity +=\
    #         unit_contributivity
    # combined_predictions = combine_predictions_list(
    #     predictions_list, index_list=best_index_list)
    # best_predictions = predictions_list[best_index_list[0]]

    # test_predictions_list = [
    #     submission_on_fold.test_predictions
    #     for submission_on_fold in selected_submissions_on_fold
    # ]
    # if any(test_predictions_list) is None:
    #     logger.error("Can't compute combined test score," +
    #                  " some submissions are untested.")
    #     combined_test_predictions = None
    #     best_test_predictions = None
    # else:
    #     combined_test_predictions = combine_predictions_list(
    #         test_predictions_list, index_list=best_index_list)
    #     best_test_predictions = test_predictions_list[best_index_list[0]]

    # return combined_predictions, best_predictions,\
    #     combined_test_predictions, best_test_predictions



import os

from sklearn.base import is_classifier
from sklearn.utils import _safe_indexing

from ..utils.importing import import_module_from_source


class SKLearnPipeline:
    """Wrapper to convert a scikit-learn estimator into a RAMP workflow.


    .. notes::

        This class should not be used directly in problem.py
        when using ramp-board database as it requires to have
        a fixed set of elements (files) in the workflow for
        the web frontend.

    Parameters
    ----------
    filename : str, default='estimator.py'
        The name of the python file used in the kit submission. In general,
        `estimator.py` is the used in RAMP starting-kit.
    addtional_filenames : list
        Typically of addtional file (e.g. external files) which will be used
        in the workflow.
    """
    def __init__(self, filename='estimator.py', additional_filenames=None):
        self.element_names = [os.path.splitext(filename)[0]]
        if additional_filenames is not None:
            self.element_names += additional_filenames
        self.filename = filename

    def train_submission(self, module_path, X, y, train_idx=None):
        """Train the estimator of a given submission.

        Parameters
        ----------
        module_path : str
            The path to the submission where `filename` is located.
        X : {array-like, sparse matrix, dataframe} of shape \
                (n_samples, n_features)
            The data matrix.
        y : array-like of shape (n_samples,)
            The target vector.
        train_idx : array-like of shape (n_training_samples,), default=None
            The training indices. By default, the full dataset will be used
            to train the model. If an array is provided, `X` and `y` will be
            subsampled using these indices.

        Returns
        -------
        estimator : estimator object
            The scikit-learn fitted on (`X`, `y`).
        """
        train_idx = slice(None, None, None) if train_idx is None else train_idx
        submission_module = import_module_from_source(
            os.path.join(module_path, self.filename),
            os.path.splitext(self.filename)[0],  # keep the module name only
            sanitize=True
        )
        estimator = submission_module.get_estimator()
        X_train = _safe_indexing(X, train_idx)
        y_train = _safe_indexing(y, train_idx)
        return estimator.fit(X_train, y_train)

    def test_submission(self, estimator_fitted, X):
        """Predict using a fitted estimator.

        Parameters
        ----------
        estimator_fitted : estimator object
            A fitted scikit-learn estimator.
        X : {array-like, sparse matrix, dataframe} of shape \
                (n_samples, n_features)
            The test data set.

        Returns
        -------
        pred : ndarray of shape (n_samples, n_classes) or (n_samples)
        """
        if is_classifier(estimator_fitted):
            return estimator_fitted.predict_proba(X)
        return estimator_fitted.predict(X)


class Estimator(SKLearnPipeline):
    """Wrapper to convert a scikit-learn estimator into a RAMP workflow.

    This workflow has 1 element that must be called `estimator.py
    """
    def __init__(self):
        super().__init__()


class EstimatorExternalData(SKLearnPipeline):
    """Wrapper to convert a scikit-learn estimator into a RAMP workflow
    in the presence of some external data.

    This workflow requires 2 elements: `estimator.py` and `external_data.csv`
    """
    def __init__(self):
        super().__init__(additional_filenames=['external_data.csv'])



from .feature_extractor_regressor import FeatureExtractorRegressor


class AirPassengers(FeatureExtractorRegressor):
    def __init__(self, workflow_element_names=[
            'feature_extractor', 'regressor', 'external_data.csv']):
        super(AirPassengers, self).__init__(workflow_element_names[:2])
        self.element_names = workflow_element_names



"""Multiclass predictions.

``y_pred`` should be two dimensional (n_samples x n_classes).

"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import numpy as np
import warnings
from .base import BasePrediction


def _multiclass_init(self, y_pred=None, y_true=None, n_samples=None):
    if y_pred is not None:
        self.y_pred = np.array(y_pred)
    elif y_true is not None:
        self._init_from_pred_labels(y_true)
    elif n_samples is not None:
        self.y_pred = np.empty((n_samples, self.n_columns), dtype=float)
        self.y_pred.fill(np.nan)
    else:
        raise ValueError(
            'Missing init argument: y_pred, y_true, or n_samples')
    self.check_y_pred_dimensions()


def _init_from_pred_labels(self, y_pred_labels):
    """Initalize y_pred to uniform for (positive) labels in y_pred_labels.

    Initialize multiclass Predictions from ground truth. y_pred_labels
    can be a single (positive) label in which case the corresponding
    column gets probability of 1.0. In the case of multilabel (k > 1
    positive labels), the columns corresponing the positive labels
    get probabilities 1/k.

    Parameters
    ----------
    y_pred_labels : list of objects or list of list of objects
        (of the same type)
    """
    type_of_label = type(self.label_names[0])
    self.y_pred = np.zeros(
        (len(y_pred_labels), len(self.label_names)), dtype=np.float64)
    for ps_i, label_list in zip(self.y_pred, y_pred_labels):
        # converting single labels to list of labels, assumed below
        if type(label_list) != np.ndarray and type(label_list) != list:
            label_list = [label_list]
        label_list = list(map(type_of_label, label_list))
        for label in label_list:
            ps_i[self.label_names.index(label)] = 1.0 / len(label_list)


@property
def _y_pred_label_index(self):
    """Multi-class y_pred is the index of the predicted label."""
    return np.argmax(self.y_pred, axis=1)


@property
def _y_pred_label(self):
    return self.label_names[self.y_pred_label_index]


@classmethod
def _combine(cls, predictions_list, index_list=None):
    if index_list is None:  # we combine the full list
        index_list = range(len(predictions_list))
    y_comb_list = np.array(
        [predictions_list[i].y_pred for i in index_list])
    # clipping probas into [0, 1], also taking care of the case of all zeros
    y_comb_list = np.clip(y_comb_list, 10 ** -15, 1 - 10 ** -15)
    # normalizing probabilities
    y_comb_list = y_comb_list / np.sum(y_comb_list, axis=2, keepdims=True)
    # I expect to see RuntimeWarnings in this block
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=RuntimeWarning)
        y_comb = np.nanmean(y_comb_list, axis=0)
    combined_predictions = cls(y_pred=y_comb)
    return combined_predictions


def make_multiclass(label_names=[]):
    Predictions = type(
        'Predictions',
        (BasePrediction,),
        {'label_names': label_names,
         'n_columns': len(label_names),
         # Multiclass ground truth is a 1D vector of labels
         'n_columns_true': 0,
         '__init__': _multiclass_init,
         '_init_from_pred_labels': _init_from_pred_labels,
         'y_pred_label_index': _y_pred_label_index,
         'y_pred_label': _y_pred_label,
         'combine': _combine,
         })
    return Predictions



from .base import BaseScoreType


class MakeCombined(BaseScoreType):
    def __init__(self, score_type, index):
        self.score_type = score_type
        self.is_lower_the_better = score_type.is_lower_the_better
        self.minimum = score_type.minimum
        self.maximum = score_type.maximum
        self.name = score_type.name
        self.precision = score_type.precision
        self.index = index

    def score_function(self, ground_truths_combined, predictions_combined,
                       valid_indexes=None):
        return self.score_type.score_function(
            ground_truths_combined.predictions_list[self.index],
            predictions_combined.predictions_list[self.index],
            valid_indexes)

    def __call__(self, y_true, y_pred):
        raise ValueError('MakeCombined score has no deep score function.')



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
back.py is a part of colored.

Copyright 2014-2017 Dimitris Zlatanidis <d.zlatanidis@gmail.com>
All rights reserved.

Colored is very simple Python library for color and formatting in terminal.

https://github.com/dslackw/colored

colored is free software: you can redistribute it and/or
modify it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""


from .colors import names


class back(object):

    ESC = '\x1b[48;5;'
    END = 'm'
    num = 0
    for color in names:
        vars()[color] = '{}{}{}'.format(ESC, num, END)
        num += 1



"""A time series feature extractor.

Train and test a time series feature extractor.

The input object is an `xarray` `Dataset`, containing possibly several
`DataArrays` corresponding to the input sequence. It contains a special burn
in period in the beginning (carried by X_ds.n_burn_in) for which we do not
give ground truth and we do not require the user to provide predictions.
The ground truth sequence `y_array` in train and the output of the user
submission `ts_fe.transform` are thus `n_burn_in` shorter than the input
sequence `X_ds`, making the training and testing slightly complicated.

The other particularity of this workflow is that the input `X_ds` that the
*test* receives may contain information about the (future) labels, so it is
technically possible to cheat. We developed a randomized technique to
safeguard against this. The workflow has two init parameters: `check_sizes`
and `check_indexs`. Both are lists of indices. The idea is that we first
run `transform` on the original `X_ds`, obtaining the feature matrix
`X_array`. Then we randomly change elements of `X_ds` after
`n_burn_in + check_index`, and then check if the features in the new
`X_check_array` change *before* `n_burn_in + check_index` wrt `X_array`.
If they do, the submission is illegal. If they don't, it is possible that
the user carefully avoided looking ahead at this prticular index, so we may
test at another index, to be added to the list `check_indexs`. The other
list `check_sizes` makes it possible to make a shorter copy of the full
sequence in this check, to save time. Obviously each `check_size` should be
bigger than the corresponding `check_index`.

"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import os

import numpy as np

from ..utils.importing import import_module_from_source


class TimeSeriesFeatureExtractor(object):
    def __init__(self, check_sizes, check_indexs, workflow_element_names=[
            'ts_feature_extractor']):
        self.element_names = workflow_element_names
        self.check_sizes = check_sizes
        self.check_indexs = check_indexs

    def train_submission(self, module_path, X_ds, y_array, train_is=None):
        """
        Train a time series feature extractor.

        `X_ds` is `n_burn_in` longer than `y_array` since `y_array` contains
        targets without the initial burn in period. `train_is` are wrt
        `y_array`, so `X_ds` has to be _extended_ by `n_burn_in` when sent to
        `ts_fe.fit`.

        """
        if train_is is None:
            # slice doesn't work here because of the way `extended_train_is`
            # is computed below
            train_is = np.arange(len(y_array))
        n_burn_in = X_ds.n_burn_in
        ts_feature_extractor = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        ts_fe = ts_feature_extractor.FeatureExtractor()
        # Fit is not required in the submissions but we add it here in case
        # of, e.g., a recurrent neural net which is impossible to train once
        # the features are digested into a classical tabular format (one row
        # per time step).
        try:
            burn_in_range = np.arange(train_is[-1], train_is[-1] + n_burn_in)
            extended_train_is = np.concatenate((train_is, burn_in_range))
            X_train_ds = X_ds.isel(time=extended_train_is)
            y_array_train = y_array[train_is]
            ts_fe.fit(X_train_ds, y_array_train)
        except AttributeError:
            pass
        return ts_fe

    def test_submission(self, trained_model, X_ds):
        """
        Test a time series feature extractor.

        `X_ds` is `n_burn_in` longer than `X_test_array` below since
        `X_test_array` contains feautures only beyond the initial burn in
        period.

        We check if the `transform` of the submission looks ahead into the
        future.

        """
        ts_fe = trained_model
        X_test_array = ts_fe.transform(X_ds)

        # Checking if feature extractor looks ahead: we change the input
        # array after index n_burn_in + check_index, and check if the first
        # check_size features have changed
        n_burn_in = X_ds.n_burn_in
        for check_size, check_index in zip(
                self.check_sizes, self.check_indexs):
            # We use a short prefix to save time
            X_check_ds = X_ds.isel(
                time=slice(0, n_burn_in + check_size)).copy(deep=True)
            # Adding random noise to future.
            # Assigning Dataset slices is not yet supported so we need to
            # iterate over the arrays. To generalize we should maybe check
            # the types.
            data_var_names = X_check_ds.data_vars.keys()
            for data_var_name in data_var_names:
                X_check_ds[data_var_name][dict(time=slice(
                    n_burn_in + check_index, None))] += np.random.normal()
            # Calling transform on changed future.
            X_check_array = ts_fe.transform(X_check_ds)
            X_neq = np.not_equal(
                X_test_array[:check_size], X_check_array[:check_size])
            x_neq = np.any(X_neq, axis=1)
            x_neq_nonzero = x_neq.nonzero()
            if len(x_neq_nonzero[0]) == 0:  # no change anywhere
                first_modified_index = check_index
            else:
                first_modified_index = np.min(x_neq_nonzero)
            # Normally, the features should not have changed before check_index
            if first_modified_index < check_index:
                message = 'The feature extractor looks into the future by' +\
                    ' at least {} time steps'.format(
                        check_index - first_modified_index)
                raise AssertionError(message)

        return X_test_array



from .scp import SCP
from .ospa import OSPA
from .precision_recall import (
    DetectionPrecision, DetectionRecall, MADCenter, MADRadius)
from .average_precision import (
    AverageDetectionPrecision, DetectionAveragePrecision)



import os
import glob
import shutil

import cloudpickle

import pytest

from rampwf.utils import import_module_from_source
from rampwf.utils.testing import (
    assert_submission, assert_notebook, blend_submissions)


PATH = os.path.dirname(__file__)


def skip_no_tensorflow():
    try:
        import tensorflow  # noqa
    except ImportError:
        return pytest.mark.skip(reason='tensorflow not available')
    return pytest.mark.basic


def _generate_grid_path_kits():
    grid = []
    for path_kit in sorted(glob.glob(os.path.join(PATH, 'kits', '*'))):
        if 'digits' in path_kit:
            grid.append(pytest.param(os.path.abspath(path_kit),
                                     marks=skip_no_tensorflow()))
        elif 'data_label' in path_kit:
            pass
        else:
            grid.append(os.path.abspath(path_kit))
    return grid


@pytest.mark.parametrize(
    "path_kit",
    _generate_grid_path_kits()
)
def test_notebook_testing(path_kit):
    # check if there is a notebook to be tested
    if len(glob.glob(os.path.join(path_kit, '*.ipynb'))):
        assert_notebook(ramp_kit_dir=path_kit)


@pytest.mark.parametrize(
    "path_kit",
    _generate_grid_path_kits()
)
def test_submission(path_kit):
    submissions = sorted(glob.glob(os.path.join(path_kit, 'submissions', '*')))
    for sub in submissions:
        # FIXME: to be removed once el-nino tests is fixed.
        if 'el_nino' in sub:
            pytest.xfail('el-nino is failing due to xarray.')
        else:
            assert_submission(
                ramp_kit_dir=path_kit,
                ramp_data_dir=path_kit,
                ramp_submission_dir=os.path.join(path_kit, 'submissions'),
                submission=os.path.basename(sub), is_pickle=True,
                save_output=False, retrain=True)


def test_blending():
    assert_submission(
        ramp_kit_dir=os.path.join(PATH, "kits", "iris"),
        ramp_data_dir=os.path.join(PATH, "kits", "iris"),
        ramp_submission_dir=os.path.join(PATH, "kits", "iris", "submissions"),
        submission='starting_kit', is_pickle=True,
        save_output=True, retrain=True)
    assert_submission(
        ramp_kit_dir=os.path.join(PATH, "kits", "iris"),
        ramp_data_dir=os.path.join(PATH, "kits", "iris"),
        ramp_submission_dir=os.path.join(PATH, "kits", "iris", "submissions"),
        submission='random_forest_10_10', is_pickle=True,
        save_output=True, retrain=True)
    blend_submissions(
        ['starting_kit', 'random_forest_10_10'],
        ramp_kit_dir=os.path.join(PATH, "kits", "iris"),
        ramp_data_dir=os.path.join(PATH, "kits", "iris"),
        ramp_submission_dir=os.path.join(PATH, "kits", "iris", "submissions"),
        save_output=True)
    # cleaning up so next test doesn't try to train "training_output"
    shutil.rmtree(os.path.join(
        PATH, "kits", "iris", "submissions", "training_output"))


def test_data_label():
    assert_submission(
        ramp_kit_dir=os.path.join(PATH, "kits", "iris_data_label"),
        ramp_data_dir=os.path.join(PATH, "kits", "iris_data_label"),
        data_label='data_label',
        ramp_submission_dir=os.path.join(
            PATH, "kits", "iris_data_label", "submissions"),
        submission='starting_kit', is_pickle=True,
        save_output=True, retrain=True)
    assert_submission(
        ramp_kit_dir=os.path.join(PATH, "kits", "iris_data_label"),
        ramp_data_dir=os.path.join(PATH, "kits", "iris_data_label"),
        data_label='data_label',
        ramp_submission_dir=os.path.join(
            PATH, "kits", "iris_data_label", "submissions"),
        submission='random_forest_10_10', is_pickle=True,
        save_output=True, retrain=True)
    blend_submissions(
        ['starting_kit', 'random_forest_10_10'],
        ramp_kit_dir=os.path.join(PATH, "kits", "iris_data_label"),
        ramp_data_dir=os.path.join(PATH, "kits", "iris_data_label"),
        data_label='data_label',
        ramp_submission_dir=os.path.join(
            PATH, "kits", "iris_data_label", "submissions"),
        save_output=True)
    # cleaning up so next test doesn't try to train "training_output"
    shutil.rmtree(os.path.join(
        PATH, "kits", "iris_data_label", "submissions", "training_output"))


def test_cloudpickle():
    """Check cloudpickle works with the way modules are imported from source.

    This only checks that an object that can be pickled with cloudpickle can
    still be pickled with cloudpickle when imported dynamically using
    import_module_from_source.
    """
    # use iris_old as the object has to be a custom class not an object
    # from a python package that is in sys.path such as a sklearn object
    kit = "iris_old"
    ramp_kit_dir = os.path.join(PATH, "kits", kit)
    ramp_data_dir = os.path.join(PATH, "kits", kit)
    ramp_submission = os.path.join(PATH, "kits", kit, "submissions",
                                   "starting_kit")

    problem_module = import_module_from_source(
        os.path.join(ramp_kit_dir, 'problem.py'), 'problem')
    workflow = problem_module.workflow
    X_train, y_train = problem_module.get_train_data(path=ramp_data_dir)
    model = workflow.train_submission(ramp_submission, X_train, y_train)

    # test cloudpickle
    cloudpickle.dumps(model)



# coding: utf-8
"""
Utility methods to print the results in a terminal using term colors
"""
import os
import platform

from pandas import option_context

from ..externals.colored import stylize, fg, attr

IS_WINDOWS = platform.system() == "Windows"
# known terminal types which can handle colors on any system
COLOR_TERMS = ['xterm-256color', 'cygwin', 'xterm-color']
# 'xterm' can handle color on macos but not on windows
IS_COLOR_TERM = 'TERM' in os.environ and (
    os.environ['TERM'] in COLOR_TERMS or (
        os.environ['TERM'] == 'xterm' and not IS_WINDOWS
    )
)

# Dictionary of term colors used for printing to terminal
fg_colors = {
    'official_train': 'light_green',
    'official_valid': 'light_blue',
    'official_test': 'red',
    'train': 'dark_sea_green_3b',
    'valid': 'light_slate_blue',
    'test': 'pink_1',
    'title': 'gold_3b',
    'warning': 'grey_46',
}


def print_title(title):
    if IS_COLOR_TERM:
        title = stylize(title, fg(fg_colors['title']) + attr('bold'))
    print(title)


def print_warning(warning):
    if IS_COLOR_TERM:
        warning = stylize(warning, fg(fg_colors['warning']))
    print(warning)


def print_df_scores(df_scores, indent=''):
    """Pretty print the scores dataframe.

    Parameters
    ----------
    df_scores : pd.DataFrame
        the score dataframe
    indent : str, default=''
        indentation if needed
    """
    with option_context("display.width", None):
        df_repr = repr(df_scores)
    df_repr_out = []
    for line, color_key in zip(df_repr.splitlines(),
                               [None, None] +
                               list(df_scores.index.values)):
        if line.strip() == 'step':
            continue
        if color_key is None:
            # table header
            if IS_COLOR_TERM:
                line = stylize(line, fg(fg_colors['title']) + attr('bold'))
        if color_key is not None:
            tokens = line.split()
            tokens_bak = tokens[:]
            if 'official_' + color_key in fg_colors:
                # line label and official score bold & bright
                if IS_COLOR_TERM:
                    label_color = fg(fg_colors['official_' + color_key])
                    tokens[0] = stylize(tokens[0], label_color + attr('bold'))
                    tokens[1] = stylize(tokens[1], label_color + attr('bold'))
            if IS_COLOR_TERM and (color_key in fg_colors):
                # other scores pale
                tokens[2:] = [stylize(token, fg(fg_colors[color_key]))
                              for token in tokens[2:]]
            for token_from, token_to in zip(tokens_bak, tokens):
                line = line.replace(token_from, token_to)
        line = indent + line
        df_repr_out.append(line)
    print('\n'.join(df_repr_out))



"""Score function for clustering efficiency.

Cluster ids should be nonnegative integers. A negative integer
will mean that the corresponding point does not belong to any
cluster.

We first identify assigned clusters by taking the max count of
unique assigned ids for each true cluster. We remove all unassigned
clusters (all assigned ids are -1) and all duplicates (the same
assigned id has majority in several true clusters) except the one
with the largest count. We add the counts, then divide by the number
of events. The score should be between 0 and 1.

"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import numpy as np
from .base import BaseScoreType


class ClusteringEfficiency(BaseScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='clustering_efficiency', precision=2):
        self.name = name
        self.precision = precision

    def __call__(self, y_true, y_pred):
        score = 0.
        event_ids = y_true[:, 0]
        y_true_cluster_ids = y_true[:, 1]
        y_pred_cluster_ids = y_pred[:, 1]

        unique_event_ids = np.unique(event_ids)
        for event_id in unique_event_ids:
            event_indices = (event_ids == event_id)
            cluster_ids_true = y_true_cluster_ids[event_indices]
            cluster_ids_pred = y_pred_cluster_ids[event_indices]

            unique_cluster_ids = np.unique(cluster_ids_true)
            n_cluster = len(unique_cluster_ids)
            n_sample = len(cluster_ids_true)

            # assigned_clusters[i] will be the predicted cluster id
            # we assign (by majority) to true cluster i
            assigned_clusters = np.empty(n_cluster, dtype='int64')
            # true_positives[i] will be the number of points in
            # predicted cluster[assigned_clusters[i]]
            true_positives = np.full(n_cluster, fill_value=0, dtype='int64')
            for i, cluster_id in enumerate(unique_cluster_ids):
                # predicted points belonging to a cluster
                found_points = cluster_ids_pred[cluster_ids_true == cluster_id]
                # nonnegative cluster_ids (negative ones are unassigned)
                assigned_points = found_points[found_points >= 0]
                # the unique nonnegative predicted cluster ids on
                # true_cluster[i]
                n_sub_cluster = len(np.unique(assigned_points))
                # We find the largest predicted cluster in the true cluster.
                if(n_sub_cluster > 0):
                    # sizes of predicted assigned cluster in true cluster[i]
                    predicted_cluster_sizes = np.bincount(
                        assigned_points.astype(dtype='int64'))
                    # If there are ties, we assign the true cluster to the
                    # predicted cluster with the smallest id (combined behavior
                    # of np.unique which sorts the ids and np.argmax which
                    # returns the first occurence of a tie).
                    assigned_clusters[i] = np.argmax(predicted_cluster_sizes)
                    true_positives[i] = len(
                        found_points[found_points == assigned_clusters[i]])
                # If none of the assigned ids are positive, the cluster is
                # unassigned and true_positive = 0
                else:
                    assigned_clusters[i] = -1
                    true_positives[i] = 0

            # resolve duplicates and count good assignments
            sorted = np.argsort(true_positives)
            true_positives_sorted = true_positives[sorted]
            assigned_clusters_sorted = assigned_clusters[sorted]
            good_clusters = assigned_clusters_sorted >= 0
            for i in range(len(assigned_clusters_sorted) - 1):
                assigned_cluster = assigned_clusters_sorted[i]
                # duplicates: only keep the last count (which is the largest
                # because of sorting)
                if assigned_cluster in assigned_clusters_sorted[i + 1:]:
                    good_clusters[i] = False
            n_good = np.sum(true_positives_sorted[good_clusters])
            score += 1. * n_good / n_sample
        score /= len(unique_event_ids)
        return score



"""Testing for regression predictions (rampwf.regression.multiclass)."""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import numpy as np
from rampwf.prediction_types import make_regression
from numpy.testing import assert_array_equal


def test_init():
    Predictions = make_regression()
    y_pred = [0.7, 0.1, 0.2]
    predictions = Predictions(y_pred=y_pred)
    assert_array_equal(predictions.y_pred, y_pred)


def test_init_empty():
    Predictions = make_regression()
    predictions = Predictions(n_samples=3)
    assert_array_equal(predictions.y_pred, np.array([np.nan, np.nan, np.nan]))
    # multi-target regression
    Predictions = make_regression(label_names=[1, 2, 3])
    predictions = Predictions(n_samples=4)
    assert_array_equal(predictions.y_pred, np.array([
        [np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan],
    ]))


def test_set_valid_in_train():
    Predictions = make_regression()
    y_pred = [0.7, 0.1, 0.2]
    predictions = Predictions(y_pred=y_pred)
    cv_y_pred = [0.6, 0.3]
    cv_predictions = Predictions(y_pred=cv_y_pred)
    updated_y_pred = [0.6, 0.1, 0.3]
    updated_predictions = Predictions(y_pred=updated_y_pred)
    predictions.set_valid_in_train(cv_predictions, [0, 2])
    assert_array_equal(predictions.y_pred, updated_predictions.y_pred)


def test_set_valid_in_train2():
    Predictions = make_regression()
    predictions = Predictions(n_samples=4)
    y_pred = [0.7, 0.1, 0.2]
    predictions_valid = Predictions(y_pred=y_pred)
    predictions.set_valid_in_train(predictions_valid, [0, 2, 3])
    assert_array_equal(predictions.y_pred, np.array([
        0.7, np.nan, 0.1, 0.2]))



"""Combined predictions.

Handling a list of Predictions.

"""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import numpy as np
from .base import BasePrediction


def _combined_init(self, y_pred=None, y_true=None, n_samples=None):
    self.predictions_list = []
    if y_pred is not None:
        start = 0
        for Predictions in self.Predictions_list:
            end = start + Predictions.n_columns
            predictions = Predictions(y_pred=y_pred[:, start:end])
            self.predictions_list.append(predictions)
            start += Predictions.n_columns
    elif y_true is not None:
        start = 0
        for Predictions in self.Predictions_list:
            n_columns = 1
            if Predictions.n_columns_true != 0:
                n_columns = Predictions.n_columns_true
            end = start + n_columns
            predictions = Predictions(y_true=y_true[:, start:end])
            self.predictions_list.append(predictions)
            start += n_columns
    elif n_samples is not None:
        for Predictions in self.Predictions_list:
            n_columns = Predictions.n_columns
            predictions = Predictions(n_samples=n_samples)
            self.predictions_list.append(predictions)
    else:
        raise ValueError(
            'Missing init argument: y_pred, y_true, or n_samples')
    self.n_columns = np.array(
        [Predictions.n_columns for Predictions in self.Predictions_list]).sum()
    self.n_columns_true = np.array(
        [Predictions.n_columns_true
         for Predictions in self.Predictions_list]).sum()


def _set_valid_in_train(self, predictions, test_is):
    """Set a cross-validation slice."""
    for to_predictions, from_predictions in zip(
            self.predictions_list, predictions.predictions_list):
        to_predictions.y_pred[test_is] = from_predictions.y_pred


@property
def _y_pred(self):
    return np.concatenate(
        [predictions.y_pred for predictions in self.predictions_list], axis=1)


def make_combined(Predictions_list):
    Predictions = type(
        'Predictions',
        (BasePrediction,),
        {'n_columns': 0,
         'n_columns_true': 0,
         'Predictions_list': Predictions_list,
         '__init__': _combined_init,
         'set_valid_in_train': _set_valid_in_train,
         'y_pred': _y_pred,
         })
    return Predictions



"""Testing for multiclass predictions (rampwf.prediction.multiclass)."""

# Author: Balazs Kegl <balazs.kegl@gmail.com>
# License: BSD 3 clause

import numpy as np
from rampwf.prediction_types import make_multiclass
from numpy.testing import assert_array_equal, assert_array_almost_equal


def test_init():
    label_names = ['Class_1', 'Class_2', 'Class_3']
    Predictions = make_multiclass(label_names=label_names)
    ps_1 = [0.7, 0.1, 0.2]
    ps_2 = [0.1, 0.8, 0.1]
    ps_3 = [0.2, 0.5, 0.3]
    predictions = Predictions(y_pred=[ps_1, ps_2, ps_3])
    assert_array_equal(predictions.y_pred, [ps_1, ps_2, ps_3])
    assert_array_equal(predictions.y_pred_label_index, [0, 1, 1])


def test_init_empty():
    label_names = ['Class_1', 'Class_2', 'Class_3']
    Predictions = make_multiclass(label_names=label_names)
    predictions = Predictions(n_samples=4)
    assert_array_equal(predictions.y_pred, np.array([
        [np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan],
        [np.nan, np.nan, np.nan],
    ]))


def test_init_from_truth():
    label_names = ['Class_1', 'Class_2', 'Class_3']
    Predictions = make_multiclass(label_names=label_names)
    y_pred_label_1 = 'Class_2'
    y_pred_label_2 = 'Class_1'
    y_pred_label_3 = 'Class_3'
    predictions = Predictions(
        y_true=[y_pred_label_1, y_pred_label_2, y_pred_label_3])
    assert_array_equal(predictions.y_pred, [[0, 1, 0], [1, 0, 0], [0, 0, 1]])
    assert_array_equal(predictions.y_pred_label_index, [1, 0, 2])


def test_list_init_from_truth():
    label_names = ['Class_1', 'Class_2', 'Class_3']
    Predictions = make_multiclass(label_names=label_names)
    y_pred_label_1 = ['Class_2']
    y_pred_label_2 = ['Class_1']
    y_pred_label_3 = ['Class_3']
    predictions = Predictions(
        y_true=[y_pred_label_1, y_pred_label_2, y_pred_label_3])
    assert_array_equal(predictions.y_pred, [[0, 1, 0], [1, 0, 0], [0, 0, 1]])
    assert_array_equal(predictions.y_pred_label_index, [1, 0, 2])


def test_multiclass_init_from_truth():
    label_names = [1, 2, 3]
    Predictions = make_multiclass(label_names=label_names)
    y_pred_label_1 = [2]
    y_pred_label_2 = [1]
    y_pred_label_3 = [3]
    predictions = Predictions(
        y_true=[y_pred_label_1, y_pred_label_2, y_pred_label_3])
    assert_array_equal(predictions.y_pred, [[0, 1, 0], [1, 0, 0], [0, 0, 1]])
    assert_array_equal(predictions.y_pred_label_index, [1, 0, 2])


def test_multiclass_init_from_labels():
    label_names = [1, 2, 3]
    Predictions = make_multiclass(label_names=label_names)
    y_pred_label_1 = 2
    y_pred_label_2 = 1
    y_pred_label_3 = 3
    predictions = Predictions(
        y_true=[y_pred_label_1, y_pred_label_2, y_pred_label_3])
    assert_array_equal(predictions.y_pred, [[0, 1, 0], [1, 0, 0], [0, 0, 1]])
    assert_array_equal(predictions.y_pred_label_index, [1, 0, 2])


def test_multilabel_init_from_truth():
    label_names = [1, 2, 3]
    Predictions = make_multiclass(label_names=label_names)
    y_pred_label_1 = [1, 3]
    y_pred_label_2 = [2]
    y_pred_label_3 = [1, 2, 3]
    predictions = Predictions(
        y_true=[y_pred_label_1, y_pred_label_2, y_pred_label_3])
    assert_array_equal(predictions.y_pred,
                       [[0.5, 0, 0.5], [0, 1, 0], [1.0 / 3, 1.0 / 3, 1.0 / 3]])
    # in case of ties, argmax returns the index of the first max
    assert_array_equal(predictions.y_pred_label_index, [0, 1, 0])


def test_set_valid_in_train():
    label_names = ['Class_1', 'Class_2', 'Class_3']
    Predictions = make_multiclass(label_names=label_names)
    predictions = Predictions(n_samples=4)
    y_pred_label_1 = ['Class_2']
    y_pred_label_2 = ['Class_1']
    y_pred_label_3 = ['Class_3']
    predictions_valid = Predictions(
        y_true=[y_pred_label_1, y_pred_label_2, y_pred_label_3])
    predictions.set_valid_in_train(predictions_valid, [0, 2, 3])
    assert_array_equal(predictions.y_pred, np.array([
        [0, 1, 0],
        [np.nan, np.nan, np.nan],
        [1, 0, 0],
        [0, 0, 1],
    ]))


def test_combine():
    label_names = ['Class_1', 'Class_2', 'Class_3']
    Predictions = make_multiclass(label_names=label_names)
    ps_11 = [0.7, 0.1, 0.2]
    ps_12 = [0.1, 0.8, 0.1]
    ps_13 = [0.2, 0.5, 0.3]
    predictions_1 = Predictions(y_pred=[ps_11, ps_12, ps_13])
    ps_21 = [0.3, 0.4, 0.3]
    ps_22 = [0.5, 0.2, 0.3]
    ps_23 = [0.1, 0.5, 0.4]
    predictions_2 = Predictions(y_pred=[ps_21, ps_22, ps_23])
    combined = Predictions.combine(
        predictions_list=[predictions_1, predictions_2],
        index_list=[0, 1])
    ps_combined_11 = [0.5, 0.25, 0.25]
    ps_combined_12 = [0.3, 0.5, 0.2]
    ps_combined_13 = [0.15, 0.5, 0.35]
    assert_array_almost_equal(
        combined.y_pred, [ps_combined_11, ps_combined_12, ps_combined_13])
    ps_31 = [-1, 0, 0.5]
    ps_32 = [-1, -1, 0]
    ps_33 = [0, 0.2, 0.3]
    predictions_3 = Predictions(y_pred=[ps_31, ps_32, ps_33])
    combined = Predictions.combine(
        predictions_list=[predictions_1, predictions_3],
        index_list=[0, 1])
    ps_combined_21 = [0.35, 0.05, 0.6]
    ps_combined_22 = [0.05 + 1. / 6, 0.4 + 1. / 6, 0.05 + 1. / 6]
    ps_combined_23 = [0.1, 0.45, 0.45]
    assert_array_almost_equal(
        combined.y_pred, [ps_combined_21, ps_combined_22, ps_combined_23])



from sklearn.metrics import f1_score
from .classifier_base import ClassifierBaseScoreType


class F1Above(ClassifierBaseScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='f1_above', precision=2, threshold=0.5):
        self.name = name
        self.precision = precision
        self.threshold = threshold

    def __call__(self, y_true, y_pred):
        f1 = f1_score(y_true, y_pred, average=None)
        return 1. * len(f1[f1 > self.threshold]) / len(f1)



"""
Extract features from a set of spatially gridded data.

This feature extractor can be used for any problem in which one has a series of
spatial grids arranged in time, y, x dimensions. The xarray library is used
to read the data into a Dataset, which is represented by X_ds.
y_array can be a set of binary labels or a regressor value. The values in
y_array are assumed to be paired exactly with the values in the first
dimension of X_ds.
"""

import os

import pandas as pd

from ..utils.importing import import_module_from_source


class GridFeatureExtractor(object):
    def __init__(self, workflow_element_names=['feature_extractor']):
        self.element_names = workflow_element_names

    def train_submission(self, module_path, X_ds, y_array, train_is=None):
        if train_is is None:
            train_is = slice(None, None, None)
        feature_extractor = import_module_from_source(
            os.path.join(module_path, self.element_names[0] + '.py'),
            self.element_names[0],
            sanitize=True
        )
        fe = feature_extractor.FeatureExtractor()
        dim_set = pd.Series(list(X_ds.dims.keys()))
        time_dim = dim_set[dim_set.str.contains("time")][0]
        fe.fit(X_ds.isel(**{time_dim: train_is}), y_array[train_is])
        return fe

    def test_submission(self, trained_model, X_df):
        fe = trained_model
        X_test_array = fe.transform(X_df)
        return X_test_array



from sklearn.metrics import roc_auc_score
from .base import BaseScoreType


class ROCAUC(BaseScoreType):
    is_lower_the_better = False
    minimum = 0.0
    maximum = 1.0

    def __init__(self, name='roc_auc', precision=2):
        self.name = name
        self.precision = precision

    def score_function(self, ground_truths, predictions, valid_indexes=None):
        """A hybrid score.

        It tests the the predicted _probability_ of the second class
        against the true _label index_ (which is 0 if the first label is the
        ground truth, and 1 if it is not, in other words, it is the
        tru probabilty of the second class). Thus we have to override the
        `Base` function here
        """
        if valid_indexes is None:
            valid_indexes = slice(None, None, None)
        y_proba = predictions.y_pred[valid_indexes][:, 1]
        y_true_proba = ground_truths.y_pred_label_index[valid_indexes]
        self.check_y_pred_dimensions(y_true_proba, y_proba)
        return self.__call__(y_true_proba, y_proba)

    def __call__(self, y_true_proba, y_proba):
        return roc_auc_score(y_true_proba, y_proba)
